from __future__ import division
from climlab.process.energy_budget import EnergyBudget
from climlab.domain.field import global_mean


class BudykoTransport(EnergyBudget):
    r"""calculates the 1 dimensional heat transport as the difference
    between the local temperature and the global mean temperature.

    :param float b:     budyko transport parameter                      \n
                        - unit: :math:`\\textrm{W} / \\left( \\textrm{m}^2 \\ ^{\circ} \\textrm{C} \\right)`   \n
                        - default value: ``3.81``

    As BudykoTransport is a :class:`~climlab.process.process.Process` it needs
    a state do be defined on. See example for details.

    **Computation Details:** \n

    In a global Energy Balance Model

    .. math::

        C \\frac{dT}{dt} = R\downarrow - R\uparrow - H

    with model state :math:`T`, the energy transport term :math:`H`
    can be described as

    .. math::

        H = b [T - \\bar{T}]

    where :math:`T` is a vector of the model temperature and :math:`\\bar{T}`
    describes the mean value of :math:`T`.

    For further information see :cite:`Budyko_1969`.

    :Example:

        Budyko Transport as a standalone process:

        .. plot:: code_input_manual/example_budyko_transport.py
           :include-source:

    """
    # implemented by m-kreuzer
    def __init__(self, b=3.81, **kwargs):
        super(BudykoTransport, self).__init__(**kwargs)
        self.b = b

    @property
    def b(self):
        r"""the budyko transport parameter in unit
        :math:`\\frac{\\textrm{W}}{\\textrm{m}^2 \\textrm{K}}`

        :getter: returns the budyko transport parameter
        :setter: sets the budyko transport parameter
        :type: float

        """
        return self._b
    @b.setter
    def b(self, value):
        self._b = value
        self.param['b'] = value

    def _compute_heating_rates(self):
        """Computes energy flux convergences to get heating rates in :math:`W/m^2`.

        """
        for varname, value in self.state.items():
            self.heating_rate[varname] = - self.b * (value - global_mean(value))



## NEED TO FIX THE PASSING OF INITIAL PARAMETERS
# especially timestep


# how easy is to implement the Stommel 1961 box model in climlab?
#  currently... it still requires a fair bit of code:
from __future__ import division
from __future__ import print_function
import numpy as np
from climlab.process.time_dependent_process import TimeDependentProcess
from climlab.domain import domain, field


box = domain.box_model_domain()
print(box.shape)

# initial condition
x = field.Field([1.,0.], domain=box)
y = field.Field([1.,1.], domain=box)
state = {'x':x, 'y':y}
# define the process
class StommelBox(TimeDependentProcess):
    def _compute(self):
        x = self.state['x']
        y = self.state['y']
        term = np.abs(-y + self.param['R']*x) / self.param['lam']
        tendencies = {}
        tendencies['y'] = (1 - y - y * term)
        tendencies['x'] = (self.param['delta'] * (1 - x) - x * term)
        return tendencies

# make a parameter dictionary
param = {'R': 2., 'lam': 1., 'delta': 1., 'timestep':0.01}
# instantiate the process
boxmodel = StommelBox(state=state, **param)
# change the timestep
boxmodel.set_timestep(num_steps_per_year=1E9)



"""Routines for calculating heat capacities for grid boxes."""

from __future__ import division
from climlab import constants as const


def atmosphere(dp):
    """Returns heat capacity of a unit area of atmosphere, in units J /m**2 / K.

    .. math::

        C_a = \\frac{c_p \\cdot dp \\cdot f_{\\textrm{mb-to-Pa}}}{g}

    where

    ==============================  ==============  ==================================================  ===============================================
    variable                        value           unit                                                description
    ==============================  ==============  ==================================================  ===============================================
    :math:`C_a`                     *output*        :math:`\\textrm{J} / \\textrm{m}^2 / \\textrm{K}`      heat capacity for atmospheric cell
    :math:`c_p`                     :math:`1004.`   :math:`\\textrm{J} / \\textrm{kg} / \\textrm{K}`       specific heat at constant pressure for dry air
    :math:`dp`                      *input*         :math:`\\textrm{mb}`                                 pressure for atmospheric cell
    :math:`f_{\\textrm{mb-to-Pa}}`   :math:`100`     :math:`\\textrm{Pa} / \\textrm{mb}`                   conversion factor from mb to Pa
    :math:`g`                       :math:`9.8`     :math:`\\textrm{m} / \\textrm{s}^2`                   gravitational acceleration
    ==============================  ==============  ==================================================  ===============================================

    **Function-call argument** \n

    :param array dp:    pressure intervals (*unit:* mb)
    :returns:           the heat capacity for atmosphere cells correspoding to
                        pressure input (*unit:* J /m**2 / K)
    :rtype:             array

    :Example:

        Calculate atmospheric heat capacity for pressure intervals of
        1, 10, 100 mb::

            >>> from climlab.utils import heat_capacity

            >>> pressure_interval = array([1,10,100]) # in mb
            >>> heat_capacity.atmosphere(pressure_interval) # in J /m**2 / K
            array([   10244.89795918,   102448.97959184,  1024489.79591837])

    """
    return const.cp * dp * const.mb_to_Pa / const.g


def ocean(dz):
    """Returns heat capacity of a unit area of water, in units J /m**2 / K.

    .. math::

        C_o = \\rho_w \\cdot c_w \\cdot dz

    where

    ==============================  ==============  ==================================================  ==================================
    variable                        value           unit                                                description
    ==============================  ==============  ==================================================  ==================================
    :math:`C_o`                     *output*        :math:`\\textrm{J} / \\textrm{m}^2 / \\textrm{K}`      heat capacity for oceanic cell
    :math:`c_w`                     :math:`4181.3`  :math:`\\textrm{J} / \\textrm{kg} / \\textrm{K}`       specific heat of liquid water
    :math:`dz`                      *input*         :math:`\\textrm{m}`                                  water depth of oceanic cell
    :math:`\\rho_w`                  :math:`1000.`   :math:`\\textrm{kg} / \\textrm{m}^3`                   density of water
    ==============================  ==============  ==================================================  ==================================

    **Function-call argument** \n

    :param array dz:    water depth of ocean cells (*unit:* m)
    :returns:           the heat capacity for ocean cells correspoding to
                        depth input (*unit:* J /m**2 / K)
    :rtype:             array

    :Example:

        Calculate atmospheric heat capacity for pressure intervals of
        1, 10, 100 m::

            >>> from climlab.utils import heat_capacity

            >>> pressure_interval = array([1,10,100]) # in m
            >>> heat_capacity.ocean(pressure_interval) # in J /m**2 / K
            array([  4.18130000e+06,   4.18130000e+07,   4.18130000e+08])

    """
    return const.rho_w * const.cw * dz


def slab_ocean(water_depth):
    """Returns heat capacity of a unit area slab of water, in units of J / m**2 / K.

    Takes input argument ``water_depth`` and calls :func:`ocean()`

    **Function-call argument** \n

    :param float:   water depth of slab ocean (*unit:* m)
    :returns:       the heat capacity for slab ocean cell (*unit:* J / m**2 / K)
    :rtype:         float

    """
    return ocean(water_depth)



r"""Solver for the 1D meridional heat diffusion equation on the sphere:

.. math::

    C\frac{\partial}{\partial t} T(\phi,t) = \frac{1}{\cos\phi} \frac{\partial}{\partial \phi} \left[ \cos\phi ~ D ~ \frac{\partial T}{\partial \phi} \right]

for a temperature state variable :math:`T(\phi,t)`,
a vertically-integrated heat capacity :math:`C`,
and arbitrary thermal diffusivity :math:`D(\phi,t)`
in units of W/m2/K.

The diffusivity :math:`D` can be a single scalar,
or optionally a vector *specified at grid cell boundaries*
(so its length must be exactly 1 greater than the length of :math:`\phi`).

:math:`D` can be modified by the user at any time
(e.g., after each timestep, if it depends on other state variables).

The heat capacity :math:`C` is normally handled automatically by CLIMLAB
as part of the grid specification.

A fully implicit timestep is used for computational efficiency. Thus the computed
tendency :math:`\frac{\partial T}{\partial t}` will depend on the timestep.

The diagnostics ``diffusive_flux`` and ``flux_convergence`` are computed
as described in the parent class ``MeridionalDiffusion``.
Two additional diagnostics are computed here,
which are meaningful if :math:`T` represents a *zonally averaged temperature*:

- ``heat_transport`` given by :math:`\mathcal{H}(\phi) = -2 \pi ~ a^2 ~ \cos\phi ~ D ~ \frac{\partial T}{\partial \phi}` in units of PW (petawatts).
- ``heat_transport_convergence`` given by :math:`-\frac{1}{2 \pi ~a^2 \cos\phi} \frac{\partial \mathcal{H}}{\partial \phi}` in units of W/m2

Non-uniform grid spacing is supported.

The state variable :math:`T` may be multi-dimensional, but the diffusion
will operate along the latitude dimension only.
"""
from __future__ import division
import numpy as np
from .meridional_advection_diffusion import MeridionalDiffusion
from climlab import constants as const


class MeridionalHeatDiffusion(MeridionalDiffusion):
    '''A 1D diffusion solver for Energy Balance Models.

    Solves the meridional heat diffusion equation

    .. math::

        C \frac{\partial T}{\partial t} = -\frac{1}{\cos\phi} \frac{\partial}{\partial \phi} \left[ -D \cos\phi \frac{\partial T}{\partial \phi} \right]

    on an evenly-spaced latitude grid, with a state variable :math:`T`,
    a heat capacity :math:`C` and diffusivity :math:`D`.

    Assuming :math:`T` is a temperature in K or degC, then the units are:

    - :math:`D` in W m-2 K-1
    - :math:`C` in J m-2 K-1

    :math:`D` is provided as input, and can be either scalar
    or vector defined at latitude boundaries.

    :math:`C` is normally handled automatically for temperature state variables in CLIMLAB.
    '''
    def __init__(self,
                 D=0.555,  # in W / m^2 / degC
                 use_banded_solver=False,
                 **kwargs):
        #  First just use a dummy value for K
        super(MeridionalHeatDiffusion, self).__init__(K=1.,
                        use_banded_solver=use_banded_solver, **kwargs)
        #  Now initialize properly
        self.D = D
        self.add_diagnostic('heat_transport', 0.*self.diffusive_flux)
        self.add_diagnostic('heat_transport_convergence', 0.*self.flux_convergence)

    @property
    def D(self):
        return self._D
    @D.setter
    def D(self, Dvalue):
        self._D = Dvalue
        self._update_diffusivity()

    def _update_diffusivity(self):
        for varname, value in self.state.items():
            heat_capacity = value.domain.heat_capacity
        # diffusivity in units of m**2/s
        self.K = self.D / heat_capacity * const.a**2

    def _update_diagnostics(self, newstate):
        super(MeridionalHeatDiffusion, self)._update_diagnostics(newstate)
        for varname, value in self.state.items():
            heat_capacity = value.domain.heat_capacity
        coslat_bounds = np.moveaxis(self._weight_bounds,-1,self.diffusion_axis_index)
        self.heat_transport[:] = (self.diffusive_flux * heat_capacity *
            2 * np.pi * const.a * coslat_bounds * 1E-15) # in PW
        self.heat_transport_convergence[:] = (self.flux_convergence *
                        heat_capacity)  # in W/m**2



from __future__ import division
from __future__ import print_function
from builtins import str
from builtins import range
from builtins import object
import numpy as np
from climlab import constants as const
from climlab.solar.orbital import OrbitalTable
from climlab.domain.field import global_mean


class OrbitalCycles(object):
    def __init__(self,
                 model,
                 kyear_start=-20.,
                 kyear_stop=0.,
                 segment_length_years=100.,
                 orbital_year_factor=1.,
                 verbose=True ):
        """Automatically integrates a process through changes in orbital parameters.

        OrbitalCycles is a module for setting up long integrations of climlab
        processes over orbital cycles.

        The duration between integration start and end time is partitioned in
        time segments over which the orbital parameters are held constant.
        The process is integrated over every time segment and the process state
        ``Ts`` is stored for each segment.

        The storage arrays are saving:

            * **current model state** at end of each segment
            * **model state averaged** over last integrated year of each segment
            * **global mean** of averaged model state over last integrated year
              of each segment

        .. note::

            Input ``kyear`` is thousands of years after present.
            For years before present, use ``kyear < 0``.


        **Initialization parameters** \n

        :param model:                       a time dependent process
        :type model:                        :class:`~climlab.process.time_dependent_process.TimeDependentProcess`
        :param float kyear_start:           integration start time.             \n
                                            As time reference
                                            is present, argument should be :math:`<0`
                                            for time before present.

                                                * *unit:* kiloyears        \n
                                                * *default value:* ``-20.``

        :param float kyear_stop:            integration stop time.             \n
                                            As time reference
                                            is present, argument should be :math:`\\le 0`
                                            for time before present.

                                                * *unit:* kiloyears        \n
                                                * *default value:* ``0.``

        :param float segment_length_years:  is the length of each integration with
                                            fixed orbital parameters. [default: 100.]
        :param float orbital_year_factor:   is an optional speed-up to the orbital cycles.
                                            [default: 1.]
        :param bool verbose:                prints product of calculation and
                                            information about computation progress
                                            [default: True]

        **Object attributes** \n

        Following object attributes are generated during initialization:

        :ivar model:                        timedependent process to be integrated
        :vartype model:                     :class:`~climlab.process.time_dependent_process.TimeDependentProcess`
        :ivar float kyear_start:            integration start time
        :ivar float kyear_stop:             integration stop time
        :ivar float segment_length_years:   length of each integration with
                                            fixed orbital parameters
        :ivar float orbital_year_factor:    speed-up factor  to the orbital cycles
        :ivar bool verbose:                 print flag
        :ivar int num_segments:             number of segments with fixed oribtal
                                            parameters, calculated through:

                                            .. math::

                                                num_{seg} = \\frac{-(kyear_{start}-kyear_{stop})*1000}{seg_{length} * orb_{factor}}

        :ivar array T_segments_global:      storage for global mean temperature
                                            for final year of each segment
        :ivar array T_segments:             storage for actual temperature at end
                                            of each segment
        :ivar array T_segments_annual:      storage for timeaveraged temperature
                                            over last year of segment \n
                                            dimension: (size(Ts), num_segments)
        :ivar array orb_kyear:              integration start time of all segments
        :ivar dict orb:                     orbital parameters for last integrated segment

        :Example:

            Integration of an energy balance model for 10,000 years with
            corresponding orbital parameters::

                from climlab.model.ebm import EBM_seasonal
                from climlab.solar.orbital_cycles import OrbitalCycles
                from climlab.surface.albedo import StepFunctionAlbedo
                ebm = EBM_seasonal()
                print ebm

                #  add an albedo feedback
                albedo = StepFunctionAlbedo(state=ebm.state, **ebm.param)
                ebm.add_subprocess('albedo', albedo)

                #  start the integration
                #  run for 10,000 orbital years, but only 1,000 model years
                experiment = OrbitalCycles(ebm, kyear_start=-20, kyear_stop=-10,
                                                        orbital_year_factor=10.)

        """
        self.model = model
        self.kyear_start = kyear_start
        self.kyear_stop = kyear_stop
        self.segment_length_years = segment_length_years
        self.orbital_year_factor = orbital_year_factor
        self.verbose = verbose
        self.num_segments = int(-(kyear_start - kyear_stop) * 1000. /
                                segment_length_years / orbital_year_factor)

        kyear_before_present = kyear_start

        if verbose:
            print("---------  OrbitalCycles  START ----------")
            print("Beginning integration for the model from " + str(kyear_start) + " to " +
                str(kyear_stop) + " kyears before present.")
            print("Integration time for each set of orbital parameters is " +
                str(segment_length_years) + " years.")
            print("Orbital cycles will be sped up by a factor " + str(orbital_year_factor))
            print("Total number of segments is " + str(self.num_segments))

        # initialize storage arrays
        self.T_segments_global = np.empty( self.num_segments )
        self.T_segments = np.empty( (self.model.Ts.size, self.num_segments) )
        self.T_segments_annual = np.empty_like( self.T_segments )
        self.orb_kyear = np.empty( self.num_segments )

        # Get orbital data table
        #orbtable = OrbitalTable()

        for n in range(self.num_segments):
            if verbose:
                print("-------------------------")
                print("Segment " + str(n) + " out of " + str(self.num_segments) )
                print( "Using orbital parameters from " + str(kyear_before_present) + " kyears before present." )
            self.orb = OrbitalTable.interp(kyear=kyear_before_present)
            #self.model.make_insolation_array( orb )
            self.model.subprocess['insolation'].orb = self.orb
            self.model.integrate_years(segment_length_years-1., verbose=False)
            #  Run one final year to characterize the current equilibrated state
            self.model.integrate_years(1.0, verbose=False)
            self.T_segments_annual[:, n] = np.squeeze(self.model.timeave['Ts'])
            self.T_segments[:, n] = np.squeeze(self.model.Ts)
            self.T_segments_global[n] = global_mean(self.model.timeave['Ts'])
            self.orb_kyear[n] = kyear_before_present
            kyear_before_present += segment_length_years / 1000. * orbital_year_factor
            if verbose:
                print( "Global mean temperature from the final year of integration is " +
                    str(self.T_segments_global[n]) + " degrees C." )
        if verbose:
            print("---------  OrbitalCycles  END ----------")



r"""Convenience classes for pre-made Energy Balance Models in CLIMLAB.

These models all solve some form of the equation

.. math::

    C \frac{\partial}{\partial t} T_s(\phi,t) = (1-\alpha)S(\phi,t) - \left[A + B T_s \right] + \frac{1}{\cos\phi} \frac{\partial}{\partial \phi} \left[ \cos\phi ~ D ~ \frac{\partial T_s}{\partial \phi} \right]

where

- :math:`\phi` is latitude
- :math:`T_s` is a zonally averaged surface temperature
- :math:`C` is a depth-integrated heat capacity
- :math:`\alpha` is an albedo (which may depend on latitude and/or temperature)
- :math:`S(\phi, t)` is the insolation
- :math:`\left[A + B T_s \right]` is a parameterization of the Outgoing Longwave Radiation to space
- the last term on the right hand side is a diffusive heat transport convergence with thermal diffusivity :math:`D` in the same units as :math:`B`

Three classes are provided, which differ in the type of insolation :math:`S`:

- ``climlab.EBM`` uses a steady idealized annual insolation (second Legendre polynomial form)
- ``climlab.EBM_annual`` uses realistic steady annual-mean insolation
- ``climlab.EBM_seasonal`` uses realistic seasonally varying insolation

The ``__init__`` method of class ``EBM`` shows how these models are assembled
from subprocesses representing each term in the above equation.


Building the Moist EBM
----------------------

There is currently no ready-made convenience class for the **moist EBM**,
but it can be readily built by swapping out the dry heat diffusion process ``climlab.dynamics.MeridionalHeatDiffusion``
with the moist equivalent ``climlab.dynamics.MeridionalMoistDiffusion``.

This sort of mixing and matching of model components is at the heart of CLIMLAB
design and functionality.

    :Example:

        .. code-block:: python

            import climlab
            #  create and display a 1D Energy Balance Model
            dry = climlab.EBM()
            print(dry)
            #  clone this model and swap out the diffusion subprocess
            moist = climlab.process_like(dry)
            diff = climlab.dynamics.MeridionalMoistDiffusion(state=moist.state, timestep=moist.timestep)
            moist.add_subprocess('diffusion', diff)
            print(moist)

We can run both models out to equilibrium and compare the results as follows:

    :Example:

        .. code-block:: python

            #  Run both models out to quasi-equilibrium
            #   print out the global mean planetary energy budget -- should be very small
            for m in [dry, moist]:
                m.integrate_years(10)
                print(climlab.global_mean(m.net_radiation))
            #  plot and compare the temperatures
            import matplotlib.pyplot as plt
            plt.figure()
            plt.plot(dry.lat, dry.Ts, label='Dry')
            plt.plot(moist.lat, moist.Ts, label='Moist')
            plt.legend()
            plt.show()
            #  plot and compare the heat transport
            plt.figure()
            plt.plot(dry.lat_bounds, dry.heat_transport, label='Dry')
            plt.plot(moist.lat_bounds, moist.heat_transport, label='Moist')
            plt.legend()
            plt.show()

"""
from __future__ import division
import numpy as np
from climlab import constants as const
from climlab.domain.field import Field, global_mean
from climlab.process import EnergyBudget, TimeDependentProcess
from climlab.utils import legendre
from climlab.domain import domain
from climlab.radiation import AplusBT, P2Insolation, AnnualMeanInsolation, DailyInsolation, SimpleAbsorbedShortwave
from climlab.surface import albedo
from climlab.dynamics import MeridionalHeatDiffusion
from climlab.domain.initial import surface_state
from scipy import integrate

#  A lot of this should be re-written / simplified
#  using more up-to-date climlab APIs for coupling processes together
#  Making sure that each subprocess properly declares inputs and diagnostics

#  For example, the basic EBM should be created with something like
#  ebm = climlab.couple([asr,olr,diff])


class EBM(TimeDependentProcess):
    """A parent class for all Energy-Balance-Model classes.

    This class sets up a typical EnergyBalance Model with following subprocesses:

    * Outgoing Longwave Radiation (OLR) parametrization through
      :class:`~climlab.radiation.AplusBT`
    * Absorbed Shortwave Radiation (ASR) through
      :class:`~climlab.radiation.SimpleAbsorbedShortwave`
    * solar insolation paramtrization through
      :class:`~climlab.radiation.P2Insolation`
    * albedo parametrization in dependence of temperature through
      :class:`~climlab.surface.StepFunctionAlbedo`
    * energy diffusion through
      :class:`~climlab.dynamics.MeridionalHeatDiffusion`

    **Initialization parameters** \n

    An instance of ``EBM`` is initialized with the following
    arguments *(for detailed information see Object attributes below)*:

    :param int num_lat:         number of equally spaced points for the
                                latitue grid. Used for domain intialization of
                                :class:`~climlab.domain.domain.zonal_mean_surface`
                                                                            \n
                                - default value: ``90``
    :param int num_lon:         number of equally spaced points in longitude
                                                                            \n
                                - default value: ``None``
    :param float S0:            solar constant                              \n
                                - unit: :math:`\\frac{\\textrm{W}}{\\textrm{m}^2}`   \n
                                - default value: ``1365.2``
    :param float A:             parameter for linear OLR parametrization
                                :class:`~climlab.radiation.AplusBT.AplusBT` \n
                                - unit: :math:`\\frac{\\textrm{W}}{\\textrm{m}^2}`   \n
                                - default value: ``210.0``
    :param float B:             parameter for linear OLR parametrization
                                :class:`~climlab.radiation.AplusBT.AplusBT` \n
                                - unit: :math:`\\frac{\\textrm{W}}{\\textrm{m}^2 \\ ^{\circ} \\textrm{C}}`   \n
                                - default value: ``2.0``
    :param float D:             diffusion parameter for Meridional Energy Diffusion
                                :class:`~climlab.dynamics.diffusion.MeridionalDiffusion`
                                                                            \n
                                - unit: :math:`\\frac{\\textrm{W}}{\\textrm{m}^2 \\ ^{\circ} \\textrm{C}}`   \n
                                - default value: ``0.555``
    :param float water_depth:   depth of :class:`~climlab.domain.domain.zonal_mean_surface`
                                domain, which the heat capacity is dependent on
                                                                            \n
                                - unit: meters                              \n
                                - default value: ``10.0``
    :param float Tf:            freezing temperature                        \n
                                - unit: :math:`^{\circ} \\textrm{C}`                    \n
                                - default value: ``-10.0``
    :param float a0:            base value for planetary albedo parametrization
                                :class:`~climlab.surface.albedo.StepFunctionAlbedo`
                                                                            \n
                                - unit: dimensionless
                                - default value: ``0.3``
    :param float a2:            parabolic value for planetary  albedo parametrization
                                :class:`~climlab.surface.albedo.StepFunctionAlbedo`
                                                                            \n
                                - unit: dimensionless
                                - default value: ``0.078``
    :param float ai:            value for ice albedo paramerization in
                                :class:`~climlab.surface.albedo.StepFunctionAlbedo`
                                                                            \n
                                - unit: dimensionless
                                - default value: ``0.62``
    :param float timestep:      specifies the EBM's timestep                \n
                                - unit: seconds
                                - default value: (365.2422 * 24 * 60 * 60 ) / 90 \n
                                  -> (90 timesteps per year)
    :param float T0:            base value for initial temperature          \n
                                - unit :math:`^{\circ} \\textrm{C}`         \n
                                - default value: ``12``
    :param float T2:            factor for 2nd Legendre polynomial
                                :class:`~climlab.utils.legendre.P2`
                                to calculate initial temperature            \n
                                - unit: dimensionless
                                - default value: ``40``




    **Object attributes** \n

    Additional to the parent class :class:`~climlab.process.EnergyBudget`
    following object attributes are generated and updated during initialization:

    :ivar dict param:       The parameter dictionary is updated with a couple
                            of the initatilzation input arguments, namely
                            ``'S0'``, ``'A'``, ``'B'``, ``'D'``, ``'Tf'``,
                            ``'water_depth'``, ``'a0'``, ``'a2'`` and ``'ai'``.
    :ivar dict domains:     If the object's ``domains`` and the ``state``
                            dictionaries are empty during initialization
                            a domain ``sfc`` is created through
                            :func:`~climlab.domain.domain.zonal_mean_surface`.
                            In the meantime the object's ``domains`` and
                            ``state`` dictionaries are updated.
    :ivar dict subprocess:  Several subprocesses are created (see above)
                            through calling
                            :func:`~climlab.process.process.Process.add_subprocess`
                            and therefore the subprocess dictionary is updated.
    :ivar bool topdown:     is set to ``False`` to call subprocess compute
                            methods first.
                            See also
                            :class:`~climlab.process.time_dependent_process.TimeDependentProcess`.
    :ivar dict diagnostics: is initialized with keys: ``'OLR'``, ``'ASR'``,
                            ``'net_radiation'``, ``'albedo'``, ``'icelat'`` and
                            ``'ice_area'`` through
                            :func:`~climlab.process.process.Process.add_diagnostic`.

    :Example:

        Creation and integration of the preconfigured Energy Balance Model::

            >>> import climlab
            >>> model = climlab.EBM()

            >>> model.integrate_years(2.)
            Integrating for 180 steps, 730.4844 days, or 2.0 years.
            Total elapsed time is 2.0 years.

        For more information how to use the EBM class, see the :ref:`Tutorial`
        chapter.

    """
    def __init__(self,
                 num_lat=90,
                 num_lon=None,
                 S0=const.S0,
                 s2=-0.48,
                 A=210.,
                 B=2.,
                 D=0.555,  # in W / m^2 / degC, same as B
                 water_depth=10.0,
                 Tf=-10.,
                 a0=0.3,
                 a2=0.078,
                 ai=0.62,
                 timestep=const.seconds_per_year/90.,
                 T0 = 12.,  # initial temperature parameters
                 T2 = -40.,  #  (2nd Legendre polynomial)
                 **kwargs):
        # Check to see if an initial state is already provided
        #  If not, make one
        if 'state' not in kwargs:
            state = surface_state(num_lat=num_lat, num_lon=num_lon,
                                  water_depth=water_depth, T0=T0, T2=T2)
            sfc = state.Ts.domain
            kwargs.update({'state': state, 'domains':{'sfc':sfc}})
        super(EBM, self).__init__(timestep=timestep, **kwargs)
        sfc = self.Ts.domain
        self.param['S0'] = S0
        self.param['s2'] = s2
        self.param['A'] = A
        self.param['B'] = B
        self.param['D'] = D
        self.param['Tf'] = Tf
        self.param['water_depth'] = water_depth
        self.param['a0'] = a0
        self.param['a2'] = a2
        self.param['ai'] = ai
        # create sub-models
        lw = AplusBT(state=self.state, **self.param)
        ins = P2Insolation(domains=sfc, **self.param)
        alb = albedo.StepFunctionAlbedo(state=self.state, **self.param)
        sw = SimpleAbsorbedShortwave(state=self.state,
                                     insolation=ins.insolation,
                                     albedo=alb.albedo,
                                     **self.param)
        diff = MeridionalHeatDiffusion(state=self.state, use_banded_solver=False, **self.param)
        self.add_subprocess('LW', lw)
        self.add_subprocess('insolation', ins)
        self.add_subprocess('albedo', alb)
        self.add_subprocess('SW', sw)
        self.add_subprocess('diffusion', diff)
        self.topdown = False  # call subprocess compute methods first
        self.add_diagnostic('net_radiation', 0.*self.Ts)

    @property
    def S0(self):
        return self.subprocess['insolation'].S0
    @S0.setter
    def S0(self, value):
        self.param['S0'] = value
        self.subprocess['insolation'].S0 = value

    def _compute(self):
        self.net_radiation = self.subprocess['SW'].ASR - self.subprocess['LW'].OLR
        return super(EBM, self)._compute()

    def global_mean_temperature(self):
        """Convenience method to compute global mean surface temperature.

        Calls :func:`~climlab.domain.field.global_mean` method which
        for the object attriute ``Ts`` which calculates the latitude weighted
        global mean of a field.

        :Example:

            Calculating the global mean temperature of initial EBM temperature::

                >>> import climlab
                >>> model = climlab.EBM(T0=14., T2=-25)

                >>> model.global_mean_temperature()
                Field(13.99873037400856)

        """
        return global_mean(self.Ts)

    def inferred_heat_transport(self):
        """Calculates the inferred heat transport by integrating the TOA
        energy imbalance from pole to pole.

        The method is calculating

        .. math::

            H(\\varphi) = 2 \pi R^2 \int_{-\pi/2}^{\\varphi} cos\phi \ R_{TOA} d\phi

        where :math:`R_{TOA}` is the net radiation at top of atmosphere.


        :return: total heat transport on the latitude grid in unit :math:`\\textrm{PW}`
        :rtype: array of size ``np.size(self.lat_lat)``

        :Example:

            .. plot:: code_input_manual/example_EBM_inferred_heat_transport.py
                :include-source:

        """
        phi = np.deg2rad(self.lat)
        energy_in = np.squeeze(self.net_radiation)
        return (1E-15 * 2 * np.math.pi * const.a**2 *
                integrate.cumtrapz(np.cos(phi)*energy_in, x=phi, initial=0.))

    # def heat_transport(self):
    #     """Returns instantaneous heat transport in unit :math:`\\textrm{PW}`
    #     on the staggered grid (bounds) through calling
    #     :func:`diffusive_heat_transport`.
    #
    #     :Example:
    #
    #         .. plot:: code_input_manual/example_EBM_heat_transport.py
    #             :include-source:
    #
    #     """
    #     return self.diffusive_heat_transport()
    #
    def diffusive_heat_transport(self):
        """Compute instantaneous diffusive heat transport in unit :math:`\\textrm{PW}`
        on the staggered grid (bounds) through calculating:

        .. math::

            H(\\varphi) = - 2 \pi R^2 cos(\\varphi) D \\frac{dT}{d\\varphi}
                        \\approx - 2 \pi R^2 cos(\\varphi) D \\frac{\Delta T}{\Delta \\varphi}

        :rtype: array of size ``np.size(self.lat_bounds)``

        THIS IS DEPRECATED AND WILL BE REMOVED IN THE FUTURE. Use the diagnostic
        ``heat_transport`` instead, which implements the same calculation.
        """
        phi = np.deg2rad(self.lat)
        phi_stag = np.deg2rad(self.lat_bounds)
        D = self.param['D']
        T = np.squeeze(self.Ts)
        dTdphi = np.diff(T) / np.diff(phi)
        dTdphi = np.append(dTdphi, 0.)
        dTdphi = np.insert(dTdphi, 0, 0.)
        return (1E-15*-2*np.math.pi*np.cos(phi_stag)*const.a**2*D*dTdphi)

    # def heat_transport_convergence(self):
    #     """Returns instantaneous convergence of heat transport.
    #
    #     .. math::
    #
    #         h(\\varphi) = - \\frac{1}{2 \pi R^2 cos(\\varphi)} \\frac{dH}{d\\varphi}
    #                     \\approx - \\frac{1}{2 \pi R^2 cos(\\varphi)} \\frac{\Delta H}{\Delta \\varphi}
    #
    #     h is the *dynamical heating rate* in unit :math:`\\textrm{W}/ \\textrm{m}^2`
    #     which is the convergence of energy transport into each latitude band,
    #     namely the difference between what's coming in and what's going out.
    #
    #     :Example:
    #
    #         .. plot:: code_input_manual/example_EBM_heat_transport_convergence.py
    #             :include-source:
    #
    #     """
    #     phi = np.deg2rad(self.lat)
    #     phi_stag = np.deg2rad(self.lat_bounds)
    #     H = 1.E15*self.heat_transport()
    #     return (-1./(2*np.math.pi*const.a**2*np.cos(phi)) *
    #             np.diff(H)/np.diff(phi_stag))


class EBM_seasonal(EBM):
    def __init__(self, a0=0.33, a2=0.25, ai=None, **kwargs):
        """A class that implements Energy Balance Models with realistic
        daily insolation.

        This class is inherited from the general :class:`~climlab.EBM`
        class and uses the insolation subprocess
        :class:`~climlab.radiation.DailyInsolation` instead of
        :class:`~climlab.radiation.P2Insolation` to compute a
        realisitc distribution of solar radiation on a daily basis.

        If argument for ice albedo ``'ai'`` is not given, the model will not
        have an albedo feedback.

        An instance of ``EBM_seasonal`` is initialized with the following
        arguments:

        :param float a0:            base value for planetary albedo parametrization
                                    :class:`~climlab.surface.albedo.StepFunctionAlbedo`
                                    [default: 0.33]
        :param float a2:            parabolic value for planetary  albedo parametrization
                                    :class:`~climlab.surface.albedo.StepFunctionAlbedo`
                                    [default: 0.25]
        :param float ai:            value for ice albedo paramerization in
                                    :class:`~climlab.surface.albedo.StepFunctionAlbedo`
                                    (optional)


        **Object attributes** \n

        Following object attributes are updated during initialization: \n

        :ivar dict param:       The parameter dictionary is updated with
                                ``'a0'`` and ``'a2'``.
        :ivar dict subprocess:  suprocess ``'insolation'`` is overwritten by
                                :class:`~climlab.radiation.insolation.DailyInsolation`.

        *if* ``'ai'`` *is not given*:

        :ivar dict param:       ``'ai'`` and ``'Tf'`` are removed from the
                                parameter dictionary (initialized by parent class
                                :class:`~climlab.model.ebm.EBM`)
        :ivar dict subprocess:  suprocess ``'albedo'`` is overwritten by
                                :class:`~climlab.surface.albedo.P2Albedo`.

        *if* ``'ai'`` *is given*:

        :ivar dict param:       The parameter dictionary is updated with
                                ``'ai'``.
        :ivar dict subprocess:  suprocess ``'albedo'`` is overwritten by
                                :class:`~climlab.surface.albedo.StepFunctionAlbedo`
                                (which basically has been there before but now is
                                updated with the new albedo parameter values).
        :Example:

            The annual distribution of solar insolation:

            .. plot:: code_input_manual/example_EBM_seasonal.py
                :include-source:

        """
        if ai is None:
            no_albedo_feedback = True
            ai = 0. # ignored but need to set a number
        else:
            no_albedo_feedback = False
        super(EBM_seasonal, self).__init__(a0=a0, a2=a2, ai=ai, **kwargs)
        self.param['a0'] = a0
        self.param['a2'] = a2
        sfc = self.domains['Ts']
        ins = DailyInsolation(domains=sfc, **self.param)
        if no_albedo_feedback:
            # Remove unused parameters here for clarity
            _ = self.param.pop('ai')
            _ = self.param.pop('Tf')
            alb = albedo.P2Albedo(domains=sfc, **self.param)
        else:
            self.param['ai'] = ai
            alb = albedo.StepFunctionAlbedo(state=self.state, **self.param)
        sw = SimpleAbsorbedShortwave(state=self.state,
                                     insolation=ins.insolation,
                                     albedo=alb.albedo,
                                     **self.param)
        self.add_subprocess('insolation', ins)
        self.add_subprocess('albedo', alb)
        self.add_subprocess('SW', sw)


class EBM_annual(EBM_seasonal):
    def __init__(self, **kwargs):
        """A class that implements Energy Balance Models with annual mean insolation.

        The annual solar distribution is calculated through averaging the
        :class:`~climlab.radiation.insolation.DailyInsolation` over time
        which has been used in used in the parent class
        :class:`~climlab.EBM_seasonal`. That is done by the subprocess
        :class:`~climlab.radiation.AnnualMeanInsolation` which is
        more realistic than the :class:`~climlab.radiation.P2Insolation`
        module used in the classical :class:`~climlab.EBM` class.

        According to the parent class :class:`~climlab.EBM_seasonal`
        the model will not have an ice-albedo feedback, if albedo ice parameter
        ``'ai'`` is not given. For details see there.


        **Object attributes** \n

        Following object attributes are updated during initialization: \n

        :ivar dict subprocess:  suprocess ``'insolation'`` is overwritten by
                                :class:`~climlab.radiation.AnnualMeanInsolation`

        :Example:

            The :class:`~climlab.EBM_annual` class uses a different
            insolation subprocess than the :class:`~climlab.EBM` class::

                >>> import climlab
                >>> model_annual = climlab.EBM_annual()

                >>> print model_annual

            .. code-block:: none
               :emphasize-lines: 9

                climlab Process of type <class 'climlab.model.ebm.EBM_annual'>.
                State variables and domain shapes:
                  Ts: (90, 1)
                The subprocess tree:
                top: <class 'climlab.EBM_annual'>
                   diffusion: <class 'climlab.dynamics.MeridionalHeatDiffusion'>
                   LW: <class 'climlab.radiation.AplusBT'>
                   albedo: <class 'climlab.surface.P2Albedo'>
                   insolation: <class 'climlab.radiation.AnnualMeanInsolation'>

        """
        super(EBM_annual, self).__init__(**kwargs)
        sfc = self.domains['Ts']
        ins = AnnualMeanInsolation(domains=sfc, **self.param)
        self.add_subprocess('insolation', ins)
        self.subprocess['SW'].insolation = ins.insolation

# an EBM that computes degree-days has an additional state variable.
#  Need to implement that
#  could make a good working example to document creating a new model class




#==============================================================================
#
# class _EBM(TimeDependentProcess):
#     def __init__(self, num_points=90, K=0.555, **kwargs):
#         # first create the model domains
#         doms = domain.zonal_mean_surface(num_points=num_points)
#         # initial surface temperature
#         lat = doms['sfc'].grid['lat'].points
#         initial = {}
#         initial['Ts'] = 12. - 40. * legendre.P2(np.sin(np.deg2rad(lat)))
#          #  Create process data structures
#         super(_EBM, self).__init__(domains=doms, state=initial, **kwargs)
#         #  first set all parameters to sensible default values
#         #self.num_points = num_points
#         # self.K = 2.2E6  # in m^2 / s
#         self.K = 0.555  # in W / m^2 / degC, same as B
#         self.A = 210.
#         self.B = 2.
#         # self.water_depth =  10.0
#         self.Tf = 0.0
#         self.S0 = const.S0
#         self.make_grid()
#         # self.albedo_noice = 0.303 + 0.0779 * P2( np.sin( self.phi ) )
#         self.albedo_noice = 0.33 + 0.25 * legendre.P2(np.sin(self.phi))
#         # self.albedo_ice = 0.62 * np.ones_like( self.phi )
#         self.albedo_ice = self.albedo_noice  # default to no albedo feedback
#         self.T = 12. - 40. * legendre.P2(np.sin(self.phi))
#         #  A dictionary of the model state variables
#         self.state = {'T': self.T}
#         self.positive_degree_days = np.zeros_like(self.phi)
#         # self.make_insolation_array()  # now called from inside set_timestep()
#         self.external_heat_source = np.zeros_like(self.phi)
#         self.set_timestep()
#
#     def make_grid(self):
#         '''Build the grid for the computation, evenly spaced in latitude.'''
#         #   dlat will be our grid spacing
#         #   lat will be our temperature grid:
#         #     an array with exactly num_points evenly spaced points
#         #   lat_stag will be a staggered grid with numpoints+1 points,
#         #     where the end points are the North and South poles
#         #  Then we convert these all to radians for the computation.
#         self.dlat = 180. / self.num_points
#         self.lat = np.linspace(-90. + self.dlat/2,
#                                90. - self.dlat/2, self.num_points)
#         self.lat_stag = np.linspace(-90., 90., self.num_points+1)
#         self.dphi = np.deg2rad(self.dlat)
#         self.phi = np.deg2rad(self.lat)
#         self.phi_stag = np.deg2rad(self.lat_stag)
#
#     def set_timestep(self, num_steps_per_year=90):
#         '''Change the timestep, given a number of steps per calendar year.'''
#         super(_EBM, self).set_timestep(num_steps_per_year)
#         self.set_water_depth()
#         self.make_insolation_array()
#
#     def set_water_depth(self, water_depth=10.):
#         '''Method for changing the water depth (heat capacity) with depth in m.
#         Also recomputes the tridiagonal diffusion matrix.'''
#         if water_depth is None:
#             try:
#                 water_depth = self.water_depth
#             except:
#                 ValueError("water_depth parameter is not specified.")
#         self.water_depth = water_depth
#         self.C = const.cw * const.rho_w * self.water_depth
#         self.delta_time_over_C = self.timestep / self.C
#         self.set_diffusivity(self.K)
#
#     def set_diffusivity(self, K=None):
#         '''Method for changing the diffusivity, with K in W/m^2/degC.
#         Recomputes the tridiagonal diffusion matrix.'''
#         if K is None:
#             try:
#                 K = self.K
#             except:
#                 ValueError("Diffusivity parameter K is not specified.")
#         self.K = K
#         self.diffTriDiag = self._make_diffusion_matrix()
#
#     def _make_diffusion_matrix(self):
#         J = self.num_points
#         # Ka = (const.cp * const.ps * const.mb_to_Pa / const.g / const.a**2 *
#         #        self.K * np.ones_like(self.phi_stag))
#         # cosKa = np.cos(self.phi_stag) * Ka
#         cosKa = np.cos(self.phi_stag) * self.K
#         Ka1 = (cosKa[0:J] / np.cos(self.phi) *
#                self.delta_time_over_C / self.dphi**2)
#         Ka3 = (cosKa[1:J+1] / np.cos(self.phi) *
#                self.delta_time_over_C / self.dphi**2)
#         Ka2 = np.insert(Ka1[1:J], 0, 0) + np.append(Ka3[0:J-1], 0)
#         #  Atmosphere tridiagonal matrix
#         diag = np.empty((3, J))
#         diag[0, 1:] = -Ka3[0:J-1]
#         diag[1, :] = 1 + Ka2
#         diag[2, 0:J-1] = -Ka1[1:J]
#         return diag
#
#     def compute_OLR(self):
#         return self.A + self.B * self.T
#
#     def make_insolation_array(self):
#         # will be overridden by daughter classes
#         raise NotImplementedError("Subclasses of _EBM must implement a method for computing insolation.")
#
#     def compute_insolation(self):
#         return self.insolation_array[:, self.day_of_year_index]
#
#     def compute_albedo(self):
#         '''Simple step-function albedo based on ice line at temperature Tf.'''
#         return np.where(self.T >= self.Tf, self.albedo_noice, self.albedo_ice)
#
#     def compute_radiation(self):
#         self.ASR = (1 - self.compute_albedo()) * self.compute_insolation()
#         self.OLR = self.compute_OLR()
#         self.net_radiation = self.ASR - self.OLR
#
#     def step_forward(self):
#         self.compute_radiation()
#         #  updated temperature due to radiation:
#         Trad = (self.T + (self.net_radiation + self.external_heat_source) *
#                 self.delta_time_over_C)
#         # Time-stepping the diffusion is just inverting this matrix problem:
#         # self.T = np.linalg.solve( self.diffTriDiag, Trad )
#         self.T = solve_banded((1, 1), self.diffTriDiag, Trad)
#         self.positive_degree_days += self.compute_degree_days()
#         super(_EBM, self).step_forward()
#
#     def compute_degree_days(self, threshold=0.):
#         """Return temperature*time in degree-days,
#         wherever temperature is above the threshold, otherwise zero."""
#         return np.where(self.T > threshold, self.T * self.timestep /
#                         const.seconds_per_day, np.zeros_like(self.T))
#
#     def do_new_calendar_year(self):
#         """This function is called once at the end of every calendar year."""
#         super(_EBM, self).do_new_calendar_year()
#         self.previous_positive_degree_days = self.positive_degree_days
#         self.positive_degree_days = np.zeros_like(self.phi)
#
#     def heat_transport(self):
#         '''Returns instantaneous heat transport in units on PW,
#         on the staggered grid.'''
#         return self.diffusive_heat_transport()
#
#     def diffusive_heat_transport( self ):
#         '''Compute instantaneous diffusive heat transport in units of PW, on the staggered grid.'''
#         #return ( 1E-15 * -2 * np.math.pi * np.cos(self.phi_stag) * const.cp * const.ps * const.mb_to_Pa / const.g * self.K *
#         #    np.append( np.append( 0., np.diff( self.T ) ), 0.) / self.dphi )
#         return ( 1E-15 * -2 * np.math.pi * np.cos(self.phi_stag) * const.a**2 * self.K *
#             np.append( np.append( 0., np.diff( self.T ) ), 0.) / self.dphi )
#
#     def heat_transport_convergence( self ):
#         '''Returns instantaneous convergence of heat transport in units of W / m^2.'''
#         return ( -1./(2*np.math.pi*const.a**2*np.cos(self.phi)) * np.diff( 1.E15*self.heat_transport() )
#             / np.diff(self.phi_stag) )
#
#     def inferred_heat_transport( self ):
#         '''Returns the inferred heat transport (in PW) by integrating the TOA energy imbalance from pole to pole.'''
#         return ( 1E-15 * 2 * np.math.pi * const.a**2 * integrate.cumtrapz( np.cos(self.phi)*self.net_radiation,
#             x=self.phi, initial=0. ) )
#
#     def find_icelines( self ):
#         '''Returns the instantaneous latitudes of any ice edges.'''
#         # This probably won't work in cases with multiple ice lines per hemisphere!
#         #  Revise!
#         iceindices = np.squeeze( np.where( self.T < self.Tf ) )
#         if iceindices.size == 0:
#             return 90.
#         elif iceindices.size == self.lat.size:
#             return 0.
#         else:
#             icelines = np.squeeze( np.where( np.diff(iceindices)>1) )
#             icelat1 = self.lat_stag[ iceindices[icelines]+1 ]
#             icelat2 = self.lat_stag[ iceindices[icelines+1] ]
#             return icelat1, icelat2
#
#     def global_mean( self, field ):
#         '''Compute the area-weighted global mean of a vector field on the latitude grid.'''
#       #return np.sum( field * np.cos( self.phi ) ) / np.sum( np.cos( self.phi ) )
#       return global_mean( field, self.phi )
#
#   def global_mean_temperature( self ):
#       '''Convenience method to compute global mean temperature.'''
#       return self.global_mean( self.T )
#==============================================================================



#==============================================================================
# class EBM_landocean( EBM_seasonal ):
#     '''A model with both land and ocean, based on North and Coakley (1979)
#     Essentially just invokes two different EBM_seasonal objects, one for ocean, one for land.
#     '''
#     def __str__(self):
#         return ( "Instance of EBM_landocean class with " +  str(self.num_points) + " latitude points." )
#
#     def __init__( self, num_points = 90 ):
#         super(EBM_landocean,self).__init__( num_points )
#         self.land_ocean_exchange_parameter = 1.0  # in W/m2/K
#
#         self.land = EBM_seasonal( num_points )
#         self.land.make_insolation_array( self.orb )
#         self.land.Tf = 0.
#         self.land.set_timestep( timestep = self.timestep )
#         self.land.set_water_depth( water_depth = 2. )
#
#         self.ocean = EBM_seasonal( num_points )
#         self.ocean.make_insolation_array( self.orb )
#         self.ocean.Tf = -2.
#         self.ocean.set_timestep( timestep = self.timestep )
#         self.ocean.set_water_depth( water_depth = 75. )
#
#         self.land_fraction = 0.3 * np.ones_like( self.land.phi )
#         self.C_ratio = self.land.water_depth / self.ocean.water_depth
#         self.T = self.zonal_mean_temperature()
#
#     def zonal_mean_temperature( self ):
#         return self.land.T * self.land_fraction + self.ocean.T * (1-self.land_fraction)
#
#     def step_forward( self ):
#         #  note.. this simple implementation is possibly problematic
#         # because the exchange should really occur simultaneously with radiation
#         # and before the implicit heat diffusion
#         self.exchange = (self.ocean.T - self.land.T) * self.land_ocean_exchange_parameter
#         self.land.step_forward()
#         self.ocean.step_forward()
#         self.land.T += self.exchange / self.land_fraction * self.land.delta_time_over_C
#         self.ocean.T -= self.exchange / (1-self.land_fraction) * self.ocean.delta_time_over_C
#         self.T = self.zonal_mean_temperature()
#         self.update_time()
#
#     #   This code should be more accurate, but it's ungainly and seems to produce just about the same result.
#     #def step_forward( self ):
#     #    self.exchange = (self.ocean.T - self.land.T) * self.land_ocean_exchange_parameter
#     #    self.land.compute_radiation( )
#     #    self.ocean.compute_radiation( )
#     #    Trad_land = ( self.land.T + ( self.land.net_radiation + self.exchange / self.land_fraction )
#     #        * self.land.delta_time_over_C )
#     #    Trad_ocean = ( self.ocean.T + ( self.ocean.net_radiation - self.exchange / (1-self.land_fraction) )
#     #        * self.ocean.delta_time_over_C )
#     #    self.land.T = solve_banded((1,1), self.land.diffTriDiag, Trad_land )
#     #    self.ocean.T = solve_banded((1,1), self.ocean.diffTriDiag, Trad_ocean )
#     #    self.T = self.zonal_mean_temperature()
#     #    self.land.update_time()
#     #    self.ocean.update_time()
#     #    self.update_time()
#
#     def integrate_years(self, years=1.0, verbose=True ):
#         #  Here we make sure that both sub-models have the current insolation.
#         self.land.make_insolation_array( self.orb )
#         self.ocean.make_insolation_array( self.orb )
#        super(EBM_landocean,self).integrate_years( years, verbose )
#==============================================================================


#  To do:
#    -  use integrated positive degree days to calculate implicit ice sheet melt potential
#    - also use these to set up a version of the model with vegetation-albedo feedback
#   - Create option to have specified extra ocean heat transport in the ocean component
#   - Create a default land-fraction that looks more like reality for the land-ocean model
#    - add diffusion of moist static energy
#  (would require re-computing the diffusion operator at each timestep, probably somewhat slower)

#==============================================================================
#
# class EBM_annual_moist( EBM_annual ):
#     def __str__(self):
#         return ( "Instance of EBM_annual_moist class with " +  str(self.num_points) + " latitude points  \n" +
#           "and global mean temperature " + str(self.global_mean_temperature()) + " degrees C.")
#
#     def __init__( self, num_points = 90 ):
#         _EBM.__init__( self, num_points )
#         self.K0 = self.K  # constant
#         self.Kperdegree = self.K0/20. # 5% increase per degree
#         self.Tref = 15.
#         self.set_diffusivity( K = self.compute_K() )
#
#     def compute_K(self):
#         # formula to compute diffusivity, linear in global mean temperature
#         return self.K0 + self.Kperdegree * (self.global_mean_temperature()-self.Tref)
#
#     def step_forward( self ):
#         # set the diffusivity, depends on global mean temperature
#         self.set_diffusivity( K = self.compute_K() )
#         _EBM.step_forward(self)
#==============================================================================



from __future__ import division
from __future__ import print_function
from builtins import str
from builtins import range
import numpy as np
import copy
from climlab import constants as const
from .process import Process
from climlab.utils import walk
from climlab.utils.attrdict import AttrDict


def couple(proclist, name='Parent'):
    #  Union of the two state dictionaries
    new_state = AttrDict()
    new_input = AttrDict()
    all_input = {}
    all_diagnotics_list = []
    timestep = const.seconds_per_year * 1E6  # very long!
    for proc in proclist:
        timestep = np.minimum(timestep, proc.timestep)
        for key in proc.state:
            new_state[key] = proc.state[key]
        all_diagnotics_list += list(proc.diagnostics.keys())
        for key in proc.input:
            all_input[key] = proc.input[key]
    for key in all_input:
        if (key not in new_state) and (key not in all_diagnotics_list):
            # This quantity is still a necessary input for the parent process
            new_input[key] = all_input[key]
    # The newly created parent process has the minimum timestep
    coupled = TimeDependentProcess(state=new_state, timestep=timestep, name=name)
    for proc in proclist:
        coupled.add_subprocess(proc.name, proc)
    for key in new_input:
        coupled.add_input(key, new_input[key])
    return coupled


class TimeDependentProcess(Process):
    """A generic parent class for all time-dependent processes.

    ``TimeDependentProcess`` is a child of the
    :class:`~climlab.process.process.Process` class and therefore inherits
    all those attributes.

    **Initialization parameters** \n

    An instance of ``TimeDependentProcess`` is initialized with the following
    arguments *(for detailed information see Object attributes below)*:

    :param float timestep:  specifies the timestep of the object (optional)
    :param str time_type:   how time-dependent-process should be computed
                            [default: 'explicit']
    :param bool topdown:    whether geneterate *process_types* in regular or
                            in reverse order [default: True]

    **Object attributes** \n

    Additional to the parent class :class:`~climlab.process.process.Process`
    following object attributes are generated during initialization:

    :ivar bool has_process_type_list:
                            information whether attribute *process_types*
                            (which is needed for :func:`compute` and build in
                            :func:`_build_process_type_list`)
                            exists or not. Attribute is set to ``'False'``
                            during initialization.
    :ivar bool topdown:     information whether the list *process_types* (which
                            contains all processes and sub-processes) should be
                            generated in regular or in reverse order.
                            See :func:`_build_process_type_list`.
    :ivar dict timeave:     a time averaged collection of all states and diagnostic
                            processes over the timeperiod that
                            :func:`integrate_years` has been called for last.
    :ivar dict tendencies:  computed difference in a timestep for each state.
                            See :func:`compute` for details.
    :ivar str time_type:    how time-dependent-process should be computed.
                            Possible values are: ``'explicit'``, ``'implicit'``,
                            ``'diagnostic'``, ``'adjustment'``.
    :ivar dict time:        a collection of all time-related attributes of the process.
                            The dictionary contains following items:

        * ``'timestep'``: see initialization parameter
        * ``'num_steps_per_year'``: see :func:`set_timestep` and :func:`timestep` for details
        * ``'day_of_year_index'``: counter how many steps have been integrated in current year
        * ``'steps'``: counter how many steps have been integrated in total
        * ``'days_elapsed'``: time counter for days
        * ``'years_elapsed'``: time counter for years
        * ``'days_of_year'``: array which holds the number of numerical steps per year, expressed in days

    """
    def __init__(self, time_type='explicit', timestep=None, topdown=True, **kwargs):
        # Create the state dataset
        self.tendencies = {}
        super(TimeDependentProcess, self).__init__(**kwargs)
        for name, var in self.state.items():
            self.tendencies[name] = var * 0.
        self.timeave = {}
        if timestep is None:
            self.set_timestep()
        else:
            self.set_timestep(timestep=timestep)
        self.time_type = time_type
        self.topdown = topdown
        self.has_process_type_list = False

    def __add__(self, other):
        newparent = couple([self,other])
        return newparent

    @property
    def timestep(self):
        """The amount of time over which :func:`step_forward` is integrating in unit seconds.

        :getter: Returns the object timestep which is stored in ``self.param['timestep']``.
        :setter: Sets the timestep to the given input. See also :func:`set_timestep`.
        :type: float

        """
        return self.param['timestep']
    @timestep.setter
    def timestep(self, value):
        num_steps_per_year = const.seconds_per_year / value
        timestep_days = value / const.seconds_per_day
        days_of_year = np.arange(0., const.days_per_year, timestep_days)
        self.time = {'timestep': value,
                     'num_steps_per_year': num_steps_per_year,
                     'day_of_year_index': 0,
                     'steps': 0,
                     'days_elapsed': 0,
                     'years_elapsed': 0,
                     'days_of_year': days_of_year,
                     'active_now': True}
        self.param['timestep'] = value

    def set_state(self, name, value):
        super(TimeDependentProcess, self).set_state(name,value)
        # Make sure that the new state variable is added to the tendencies dict
        self.tendencies[name] = value * 0.

    def set_timestep(self, timestep=const.seconds_per_day, num_steps_per_year=None):
        """Calculates the timestep in unit seconds
        and calls the setter function of :func:`timestep`

        :param float timestep:              the amount of time over which
                                            :func:`step_forward` is integrating
                                            in unit seconds [default: 24*60*60]
        :param float num_steps_per_year:    a number of steps per calendar year
                                            (optional)

        If the parameter *num_steps_per_year* is specified and not ``None``,
        the timestep is calculated accordingly and therefore the given input
        parameter *timestep* is ignored.

        """
        if num_steps_per_year is not None:
            timestep = const.seconds_per_year / num_steps_per_year
        # Need a more sensible approach for annual cycle stuff
        self.timestep = timestep

    def compute(self):
        """Computes the tendencies for all state variables given current state
        and specified input.

        The function first computes all diagnostic processes. They don't produce
        any tendencies directly but they may affect the other processes (such as
        change in solar distribution). Subsequently, all tendencies and
        diagnostics for all explicit processes are computed.

        Tendencies due to implicit and adjustment processes need to be
        calculated from a state that is already adjusted after explicit
        alteration. For that reason the explicit tendencies are applied to the
        states temporarily. Now all tendencies from implicit processes are
        calculated by matrix inversions and similar to the explicit tendencies,
        the implicit ones are applied to the states temporarily. Subsequently,
        all instantaneous adjustments are computed.

        Then the changes that were made to the states from explicit and implicit
        processes are removed again as this
        :class:`~climlab.process.time_dependent_process.TimeDependentProcess.compute()`
        function is supposed to calculate only tendencies and not apply them
        to the states.

        Finally, all calculated tendencies from all processes are collected
        for each state, summed up and stored in the dictionary
        ``self.tendencies``, which is an attribute of the time-dependent-process
        object, for which the
        :class:`~climlab.process.time_dependent_process.TimeDependentProcess.compute()`
        method has been called.


        **Object attributes** \n

        During method execution following object attributes are modified:

        :ivar dict tendencies:  dictionary that holds tendencies for all states
                                is calculated for current timestep through
                                adding up tendencies from explicit, implicit and
                                adjustment processes.
        :ivar dict diagnostics: process diagnostic dictionary is updated
                                by diagnostic dictionaries of subprocesses
                                after computation of tendencies.

        """
        #  First reset tendencies to zero -- recomputing them is the point of this method
        for varname in self.tendencies:
            self.tendencies[varname] *= 0.
        if not self.has_process_type_list:
            self._build_process_type_list()
        tendencies = {}
        ignored = self._compute_type('diagnostic')
        tendencies['explicit'] = self._compute_type('explicit')
        #  Tendencies due to implicit and adjustment processes need to be
        #  calculated from a state that is already adjusted after explicit stuff
        #  So apply the tendencies temporarily and then remove them again
        for name, var in self.state.items():
            var += tendencies['explicit'][name] * self.timestep
        # Now compute all implicit processes -- matrix inversions
        tendencies['implicit'] = self._compute_type('implicit')
        #  Same deal ... temporarily apply tendencies from implicit step
        for name, var in self.state.items():
            var += tendencies['implicit'][name] * self.timestep
        # Finally compute all instantaneous adjustments -- expressed as explicit forward step
        tendencies['adjustment'] = self._compute_type('adjustment')
        #  Now remove the changes from the model state
        for name, var in self.state.items():
            var -= ( (tendencies['implicit'][name] + tendencies['explicit'][name]) *
                    self.timestep)
        #  Sum up all subprocess tendencies
        for proctype in ['explicit', 'implicit', 'adjustment']:
            for varname, tend in tendencies[proctype].items():
                self.tendencies[varname] += tend
        # Finally compute my own tendencies, if any
        self_tend = self._compute()
        #  Adjustment processes _compute method returns absolute adjustment
        #  Needs to be converted to rate of change
        if self.time_type == 'adjustment':
            for varname, adj in self_tend.items():
                self_tend[varname] /= self.timestep
        for varname, tend in self_tend.items():
            self.tendencies[varname] += tend
        return self.tendencies

    def _compute_type(self, proctype):
        """Computes tendencies due to all subprocesses of given type
        ``'proctype'``. Also pass all diagnostics up to parent process."""
        tendencies = {}
        for varname in self.state:
            tendencies[varname] = 0. * self.state[varname]
        for proc in self.process_types[proctype]:
            #  Asynchronous coupling
            #  if subprocess has longer timestep than parent
            #  We compute subprocess tendencies once
            #   and apply the same tendency at each substep
            step_ratio = int(proc.timestep / self.timestep)
            #  Does the number of parent steps divide evenly by the ratio?
            #  If so, it's time to do a subprocess step.
            if self.time['steps'] % step_ratio == 0:
                proc.time['active_now'] = True
                tenddict = proc.compute()
            else:
                # proc.tendencies is unchanged from last subprocess timestep if we didn't recompute it above
                proc.time['active_now'] = False
                tenddict = proc.tendencies
            for name, tend in tenddict.items():
                tendencies[name] += tend
            for diagname, value in proc.diagnostics.items():
                self.__setattr__(diagname, value)
        return tendencies

    def _compute(self):
        """Where the tendencies are actually computed...

        Needs to be implemented for each daughter class

        Returns a dictionary with same keys as self.state"""
        tendencies = {}
        for name, value in self.state.items():
            tendencies[name] = value * 0.
        return tendencies

    def _build_process_type_list(self):
        """Generates lists of processes organized by process type.

        Following object attributes are generated or updated:

        :ivar dict process_types:   a dictionary with entries:
                                    ``'diagnostic'``, ``'explicit'``,
                                    ``'implicit'`` and ``'adjustment'`` which
                                    point to a list of processes according to
                                    the process types.

        The ``process_types`` dictionary is created while walking
        through the processes with :func:`~climlab.utils.walk.walk_processes`

        CHANGING THIS TO REFER ONLY TO THE CURRENT LEVEL IN SUBPROCESS TREE

        """
        self.process_types = {'diagnostic': [], 'explicit': [], 'implicit': [], 'adjustment': []}
        #for name, proc, level in walk.walk_processes(self, topdown=self.topdown):
        #    self.process_types[proc.time_type].append(proc)
        for name, proc in self.subprocess.items():
            self.process_types[proc.time_type].append(proc)
        self.has_process_type_list = True

    def step_forward(self):
        """Updates state variables with computed tendencies.

        Calls the :func:`compute` method to get current tendencies for all
        process states. Multiplied with the timestep and added up to the state
        variables is updating all model states.

        :Example:

            ::

                >>> import climlab
                >>> model = climlab.EBM()

                >>> # checking time step counter
                >>> model.time['steps']
                0

                >>> # stepping the model forward
                >>> model.step_forward()

                >>> # step counter increased
                >>> model.time['steps']
                1

        """
        tenddict = self.compute()
        #  Total tendency is applied as an explicit forward timestep
        # (already accounting properly for order of operations in compute() )
        for varname, tend in tenddict.items():
            self.state[varname] += tend * self.timestep
        # Update all time counters for this and all subprocesses in the tree
        #  Also pass diagnostics up the process tree
        for name, proc, level in walk.walk_processes(self, ignoreFlag=True):
            if proc.time['active_now']:
                proc._update_time()

    def compute_diagnostics(self, num_iter=3):
        """Compute all tendencies and diagnostics, but don't update model state.
        By default it will call compute() 3 times to make sure all
        subprocess coupling is accounted for. The number of iterations can
        be changed with the input argument.

        """
        for n in range(num_iter):
            ignored = self.compute()

    def _update_time(self):
        """Increments the timestep counter by one.

        Furthermore ``self.time['days_elapsed']`` and
        ``self.time['num_steps_per_year']`` are updated.

        The function is called by the time stepping methods.

        """
        self.time['steps'] += 1
        # time in days since beginning
        self.time['days_elapsed'] += self.time['timestep'] / const.seconds_per_day
        if self.time['day_of_year_index'] >= self.time['num_steps_per_year']-1:
            self._do_new_calendar_year()
        else:
            self.time['day_of_year_index'] += 1

    def _do_new_calendar_year(self):
        """This function is called once at the end of every calendar year.

        It updates ``self.time['years_elapsed']`` and
        ``self.time['day_of_year_index']``
        """
        self.time['day_of_year_index'] = 0  # back to Jan. 1
        self.time['years_elapsed'] += 1

    def integrate_years(self, years=1.0, verbose=True):
        """Integrates the model by a given number of years.

        :param float years:     integration time for the model in years
                                [default: 1.0]
        :param bool verbose:    information whether model time details
                                should be printed [default: True]

        It calls :func:`step_forward` repetitively and calculates a time
        averaged value over the integrated period for every model state and all
        diagnostics processes.

        :Example:

            ::

                >>> import climlab
                >>> model = climlab.EBM()

                >>> model.global_mean_temperature()
                Field(11.997968598413685)

                >>> model.integrate_years(2.)
                Integrating for 180 steps, 730.4844 days, or 2.0 years.
                Total elapsed time is 2.0 years.

                >>> model.global_mean_temperature()
                Field(13.531055349437258)


        """
        days = years * const.days_per_year
        numsteps = int(self.time['num_steps_per_year'] * years)
        if verbose:
            print("Integrating for " + str(numsteps) + " steps, "
                  + str(days) + " days, or " + str(years) + " years.")
        #  begin time loop
        for count in range(numsteps):
            # Compute the timestep
            self.step_forward()
            if count == 0:
                # on first step only...
                #  This implements a generic time-averaging feature
                # using the list of model state variables
                self.timeave = self.state.copy()
                # add any new diagnostics to the timeave dictionary
                self.timeave.update(self.diagnostics)
                # reset all values to zero
                for varname, value in self.timeave.items():
                # moves on to the next varname if value is None
                # this preserves NoneType diagnostics
                    if value is None:
                        continue
                    self.timeave[varname] = 0*value
            # adding up all values for each timestep
            for varname in list(self.timeave.keys()):
                try:
                    self.timeave[varname] += self.state[varname]
                except:
                    try:
                        self.timeave[varname] += self.diagnostics[varname]
                    except: pass
        # calculating mean values through dividing the sum by number of steps
        for varname, value in self.timeave.items():
            if value is None:
                continue
            self.timeave[varname] /= numsteps
        if verbose:
            print("Total elapsed time is %s years."
                  % str(self.time['days_elapsed']/const.days_per_year))

    def integrate_days(self, days=1.0, verbose=True):
        """Integrates the model forward for a specified number of days.

        It convertes the given number of days into years and calls
        :func:`integrate_years`.

        :param float days:      integration time for the model in days
                                [default: 1.0]
        :param bool verbose:    information whether model time details
                                should be printed [default: True]

        :Example:

            ::

                >>> import climlab
                >>> model = climlab.EBM()

                >>> model.global_mean_temperature()
                Field(11.997968598413685)

                >>> model.integrate_days(80.)
                Integrating for 19 steps, 80.0 days, or 0.219032740466 years.
                Total elapsed time is 0.211111111111 years.

                >>> model.global_mean_temperature()
                Field(11.873680783355553)

        """
        years = days / const.days_per_year
        self.integrate_years(years=years, verbose=verbose)

    def integrate_converge(self, crit=1e-4, verbose=True):
        """Integrates the model until model states are converging.

        :param crit:            exit criteria for difference of iterated
                                solutions [default: 0.0001]
        :type crit:             float
        :param bool verbose:    information whether total elapsed time
                                should be printed [default: True]

        :Example:

            ::

                >>> import climlab
                >>> model = climlab.EBM()

                >>> model.global_mean_temperature()
                Field(11.997968598413685)

                >>> model.integrate_converge()
                Total elapsed time is 10.0 years.

                >>> model.global_mean_temperature()
                Field(14.288155406577301)

        """
        # implemented by m-kreuzer
        for varname, value in self.state.items():
            value_old = copy.deepcopy(value)
            self.integrate_years(1,verbose=False)
            while np.max(np.abs(value_old-value)) > crit :
                value_old = copy.deepcopy(value)
                self.integrate_years(1,verbose=False)
        if verbose == True:
            print("Total elapsed time is %s years."
                  % str(self.time['days_elapsed']/const.days_per_year))



'''
climlab wrappers for the NCAR CAM3 radiation code

Input arguments and diagnostics follow specifications in
:class:`~climlab.radiation._Radiation`

    :Example:

        Here is a quick example of setting up a single-column
        Radiative-Convective model with fixed relative humdity::

            import climlab
            alb = 0.25
            #  State variables (Air and surface temperature)
            state = climlab.column_state(num_lev=30)
            #  Parent model process
            rcm = climlab.TimeDependentProcess(state=state)
            #  Fixed relative humidity
            h2o = climlab.radiation.ManabeWaterVapor(state=state)
            #  Couple water vapor to radiation
            rad = climlab.radiation.CAM3(state=state, specific_humidity=h2o.q, albedo=alb)
            #  Convective adjustment
            conv = climlab.convection.ConvectiveAdjustment(state=state, adj_lapse_rate=6.5)
            #  Couple everything together
            rcm.add_subprocess('Radiation', rad)
            rcm.add_subprocess('WaterVapor', h2o)
            rcm.add_subprocess('Convection', conv)
            #  Run the model
            rcm.integrate_years(1)
            #  Check for energy balance
            print rcm.ASR - rcm.OLR


'''
from __future__ import division, print_function, absolute_import
import numpy as np
from climlab import constants as const
from climlab.utils.thermo import vmr_to_mmr
from climlab.radiation.radiation import _Radiation_SW, _Radiation_LW
import warnings

#  Wrapping these imports in try/except to avoid failures during documentation building on readthedocs
try:
    import climlab_cam3_radiation as _cam3
except Exception:
    warnings.warn('Cannot import and initialize compiled Fortran extension, CAM3 module will not be functional.')


class CAM3(_Radiation_SW, _Radiation_LW):
    '''
    climlab wrapper for the CAM3 radiation code.

    For some details about inputs and diagnostics, see the `radiation` module.
    '''
    def __init__(self,
                 **kwargs):
        super(CAM3, self).__init__(**kwargs)

        self.KM = self.lev.size
        try:
            self.JM = self.lat.size
        except:
            self.JM = 1
        try:
            self.IM = self.lon.size
        except:
            self.IM = 1
        self.do_sw = 1  # '1=do, 0=do not compute SW'
        self.do_lw = 1  # '1=do, 0=do not compute LW'
        self.in_cld = 0 # '1=in-cloud, 0=grid avg cloud water path'

    def _climlab_to_cam3(self, field):
        '''Prepare field with proper dimension order.
        CAM3 code expects 3D arrays with (KM, JM, 1)
        and 2D arrays with (JM, 1).

        climlab grid dimensions are any of:
            - (KM,)
            - (JM, KM)
            - (JM, IM, KM)

        (longitude dimension IM not yet implemented here).'''
        if np.isscalar(field):
            return field
        #  Check to see if column vector needs to be replicated over latitude
        elif self.JM > 1:
            if (field.shape == (self.KM,)):
                return np.tile(field[...,np.newaxis], self.JM)
            else:
                return np.squeeze(np.transpose(field))[..., np.newaxis]
        else:  #  1D vertical model
            return field[..., np.newaxis, np.newaxis]

    def _cam3_to_climlab(self, field):
        ''' Output is either (KM, JM, 1) or (JM, 1).
        Transform this to...
            - (KM,) or (1,)  if JM==1
            - (KM, JM) or (JM, 1)   if JM>1

        (longitude dimension IM not yet implemented).'''
        if self.JM > 1:
            if len(field.shape)==2:
                return field
            elif len(field.shape)==3:
                return np.squeeze(np.transpose(field))
        else:
            return np.squeeze(field)

    def _prepare_arguments(self):
        # scalar integer arguments
        KM = self.KM
        JM = self.JM
        IM = self.IM
        do_sw = self.do_sw
        do_lw = self.do_lw
        in_cld = self.in_cld
        # scalar real arguments
        g = const.g
        Cpd = const.cp
        epsilon = const.Rd / const.Rv
        stebol = const.sigma
        scon = self.S0
        eccf = self.eccentricity_factor
        #  Well-mixed greenhouse gases -- scalar values
        CO2vmr = self.absorber_vmr['CO2']
        N2Ovmr = self.absorber_vmr['N2O']
        CH4vmr = self.absorber_vmr['CH4']
        CFC11vmr = self.absorber_vmr['CFC11']
        CFC12vmr = self.absorber_vmr['CFC12']
        # array input
        Tatm = self._climlab_to_cam3(self.Tatm)
        Ts = self._climlab_to_cam3(self.Ts)
        #insolation = self._climlab_to_cam3(self.insolation * np.ones_like(self.Ts))
        coszen = self._climlab_to_cam3(self.coszen * np.ones_like(self.Ts))
        aldif = self._climlab_to_cam3(self.aldif * np.ones_like(self.Ts))
        aldir = self._climlab_to_cam3(self.aldir * np.ones_like(self.Ts))
        asdif = self._climlab_to_cam3(self.asdif * np.ones_like(self.Ts))
        asdir = self._climlab_to_cam3(self.asdir * np.ones_like(self.Ts))
        #  surface pressure should correspond to model domain!
        ps = self._climlab_to_cam3(self.lev_bounds[-1] * np.ones_like(self.Ts))
        p = self._climlab_to_cam3(self.lev * np.ones_like(self.Tatm))
        #   why are we passing missing instead of the actual layer thicknesses?
        dp = np.zeros_like(p) - 99. # set as missing
        #dp = np.diff(self.lev_bounds)
        #dp = self._climlab_to_cam3(dp * np.ones_like(self.Tatm))
        # Surface upwelling LW
        flus = self._climlab_to_cam3(np.zeros_like(self.Ts) - 99.) # set to missing as default
        # spatially varying gases
        q = self._climlab_to_cam3(self.specific_humidity * np.ones_like(self.Tatm))
        O3vmr = self._climlab_to_cam3(self.absorber_vmr['O3'] * np.ones_like(self.Tatm))
        # convert to mass mixing ratio (needed by CAM3 driver)
        #  The conversion factor is m_o3 / m_air = 48.0 g/mol / 28.97 g/mol
        O3mmr = vmr_to_mmr(O3vmr, gas='O3')
        # cloud fields
        cldfrac = self._climlab_to_cam3(self.cldfrac * np.ones_like(self.Tatm))
        clwp = self._climlab_to_cam3(self.clwp * np.ones_like(self.Tatm))
        ciwp = self._climlab_to_cam3(self.ciwp * np.ones_like(self.Tatm))
        r_liq = self._climlab_to_cam3(self.r_liq * np.ones_like(self.Tatm))
        r_ice = self._climlab_to_cam3(self.r_ice * np.ones_like(self.Tatm))
        #  The ordered list of input fields needed by the CAM3 driver
        args = [KM, JM, IM, do_sw, do_lw, p, dp, ps, Tatm, Ts,
                q, O3mmr, cldfrac, clwp, ciwp, in_cld,
                aldif, aldir, asdif, asdir, eccf, coszen,
                scon, flus, r_liq, r_ice,
                CO2vmr, N2Ovmr, CH4vmr, CFC11vmr,
                CFC12vmr, g, Cpd, epsilon, stebol]
        return args

    def _compute_heating_rates(self):
        # List of arguments to be passed to extension
        args = self._prepare_arguments()
        (TdotRad, SrfRadFlx, qrs, qrl, swflx, swflxc, lwflx, lwflxc, SwToaCf,
            SwSrfCf, LwToaCf, LwSrfCf, LwToa, LwSrf, SwToa, SwSrf,
            swuflx, swdflx, swuflxc, swdflxc,
            lwuflx, lwdflx, lwuflxc, lwdflxc) = _cam3.driver(*args)
        # most of these output fields are unnecessary here
        #  we compute everything from the up and downwelling fluxes
        #  Should probably simplify the fortran wrapper
        #  fluxes at layer interfaces
        self.LW_flux_up = self._cam3_to_climlab(lwuflx) + 0.*self.LW_flux_up
        self.LW_flux_down = self._cam3_to_climlab(lwdflx) + 0.*self.LW_flux_down
        self.LW_flux_up_clr = self._cam3_to_climlab(lwuflxc) + 0.*self.LW_flux_up_clr
        self.LW_flux_down_clr = self._cam3_to_climlab(lwdflxc) + 0.*self.LW_flux_down_clr
        #  fluxes at layer interfaces
        self.SW_flux_up = self._cam3_to_climlab(swuflx) + 0.*self.SW_flux_up
        self.SW_flux_down = self._cam3_to_climlab(swdflx) + 0.*self.SW_flux_down
        self.SW_flux_up_clr = self._cam3_to_climlab(swuflxc) + 0.*self.SW_flux_up_clr
        self.SW_flux_down_clr = self._cam3_to_climlab(swdflxc) + 0.*self.SW_flux_down_clr
        #  Compute quantities derived from fluxes
        self._compute_SW_flux_diagnostics()
        self._compute_LW_flux_diagnostics()
        #  calculate heating rates from flux divergence
        #   this is the total UPWARD flux
        total_flux = self.LW_flux_net - self.SW_flux_net
        LWheating_Wm2 = np.array(np.diff(self.LW_flux_net, axis=-1)) + 0.*self.Tatm
        LWheating_clr_Wm2 = np.array(np.diff(self.LW_flux_net_clr, axis=-1)) + 0.*self.Tatm
        SWheating_Wm2 = np.array(-np.diff(self.SW_flux_net, axis=-1)) + 0.*self.Tatm
        SWheating_clr_Wm2 = np.array(-np.diff(self.SW_flux_net_clr, axis=-1)) + 0.*self.Tatm
        self.heating_rate['Ts'] = np.array(-total_flux[..., -1, np.newaxis]) + 0.*self.Ts
        self.heating_rate['Tatm'] = LWheating_Wm2 + SWheating_Wm2
        #  Convert to K / day
        Catm = self.Tatm.domain.heat_capacity
        self.TdotLW = LWheating_Wm2 / Catm * const.seconds_per_day
        self.TdotLW_clr = LWheating_clr_Wm2 / Catm * const.seconds_per_day
        self.TdotSW = SWheating_Wm2 / Catm * const.seconds_per_day
        self.TdotSW_clr = SWheating_clr_Wm2 / Catm * const.seconds_per_day


class CAM3_LW(CAM3):
    def __init__(self, **kwargs):
        super(CAM3_LW, self).__init__(**kwargs)
        self.do_sw = 0  # '1=do, 0=do not compute SW'
        self.do_lw = 1  # '1=do, 0=do not compute LW'
        #  Albedo needs to be set to 1 currently, otherwise
        #  the CAM code computes solar heating of surface.
        self.asdif = 0.*self.Ts + 1.
        self.asdir = 0.*self.Ts + 1.
        self.aldif = 0.*self.Ts + 1.
        self.aldir = 0.*self.Ts + 1.


class CAM3_SW(CAM3):
    def __init__(self, **kwargs):
        super(CAM3_SW, self).__init__(**kwargs)
        self.do_sw = 1  # '1=do, 0=do not compute SW'
        self.do_lw = 0  # '1=do, 0=do not compute LW'

##  Better to seperate out the LW and SW schemes into seperate Fortran drivers,
## like for RRTM. Among other things, this will help avoid name conflicts.



from __future__ import division
import numpy as np
import climlab
import pytest
from climlab.radiation.rrtm import _climlab_to_rrtm, _rrtm_to_climlab
from climlab.tests.xarray_test import to_xarray

num_lev = 30

@pytest.mark.compiled
@pytest.mark.fast
def test_rrtmg_lw_creation():
    state = climlab.column_state(num_lev=num_lev, water_depth=5.)
    rad = climlab.radiation.RRTMG_LW(state=state)
    #  are the transformations reversible?
    assert np.all(_rrtm_to_climlab(_climlab_to_rrtm(rad.Ts)) == rad.Ts)
    assert np.all(_rrtm_to_climlab(_climlab_to_rrtm(rad.Tatm)) == rad.Tatm)

@pytest.mark.compiled
@pytest.mark.fast
def test_rrtm_creation():
    # initial state (temperatures)
    state = climlab.column_state(num_lev=num_lev, num_lat=1, water_depth=5.)
    #  Create a RRTM radiation model
    rad = climlab.radiation.RRTMG(state=state)
    rad.step_forward()
    assert type(rad.subprocess['LW']) is climlab.radiation.RRTMG_LW
    assert type(rad.subprocess['SW']) is climlab.radiation.RRTMG_SW
    assert hasattr(rad, 'OLR')
    assert hasattr(rad, 'OLRclr')
    assert hasattr(rad, 'ASR')
    assert hasattr(rad, 'ASRclr')
    # Test the xarray interface
    to_xarray(rad)

@pytest.mark.compiled
@pytest.mark.fast
def test_swap_component():
    # initial state (temperatures)
    state = climlab.column_state(num_lev=num_lev, num_lat=1, water_depth=5.)
    #  Create a RRTM radiation model
    rad = climlab.radiation.RRTMG(state=state)
    rad.step_forward()
    #  Swap out the longwave model for CAM3
    rad.remove_subprocess('LW')
    rad.step_forward()
    rad.add_subprocess('LW', climlab.radiation.CAM3_LW(state=state))
    rad.step_forward()
    assert hasattr(rad, 'OLR')

@pytest.mark.compiled
@pytest.mark.fast
def test_multidim():
    state = climlab.column_state(num_lev=40, num_lat=3, water_depth=5.)
    rad = climlab.radiation.RRTMG_LW(state=state)
    #  are the transformations reversible?
    assert np.all(_rrtm_to_climlab(_climlab_to_rrtm(rad.Ts)) == rad.Ts)
    assert np.all(_rrtm_to_climlab(_climlab_to_rrtm(rad.Tatm)) == rad.Tatm)
    # Can we integrate?
    rad.step_forward()
    assert rad.OLR.shape == rad.Ts.shape

@pytest.mark.compiled
@pytest.mark.fast
def test_cloud():
    '''Put a high cloud layer in a radiative model.
    The all-sky ASR should be lower than clear-sky ASR.
    The all-sky OLR should be lower than clear-sky OLR.'''
    #  State variables (Air and surface temperature)
    state = climlab.column_state(num_lev=50, water_depth=1.)
    lev = state.Tatm.domain.axes['lev'].points
    #  Define some local cloud characteristics
    cldfrac = 0.5  # layer cloud fraction
    r_liq = 14.  # Cloud water drop effective radius (microns)
    clwp = 60.  # in-cloud liquid water path (g/m2)
    #  The cloud fraction is a Gaussian bump centered at level i
    i = 25
    mycloud = {'cldfrac': cldfrac*np.exp(-(lev-lev[i])**2/(2*25.)**2),
               'clwp': np.zeros_like(state.Tatm) + clwp,
               'r_liq': np.zeros_like(state.Tatm) + r_liq,}
    #  Test both RRTMG and CAM3:
    #for module in [climlab.radiation.RRTMG, climlab.radiation.CAM3]:
    #  Apparently clouds in CAM3 are not working. Save this for later
    for module in [climlab.radiation.RRTMG]:
        rad = module(state=state, **mycloud)
        rad.compute_diagnostics()
        assert(rad.ASR - rad.ASRclr < 0.)
        assert(rad.OLR - rad.OLRclr < 0.)

@pytest.mark.compiled
@pytest.mark.slow
def test_radiative_forcing():
    '''Run a single-column radiative-convective model with RRTMG radiation
    out to equilibrium. Clone the model, double CO2 and measure the instantaneous
    change in TOA flux. It should be positive net downward flux.'''
    #  State variables (Air and surface temperature)
    state = climlab.column_state(num_lev=30, water_depth=1.)
    #  Fixed relative humidity
    h2o = climlab.radiation.ManabeWaterVapor(name='WaterVapor', state=state)
    #  Couple water vapor to radiation
    #   Set icld=0 for clear-sky only (no need to call cloud overlap routine)
    rad = climlab.radiation.RRTMG(name='Radiation',
                                  state=state,
                                  specific_humidity=h2o.q,
                                  icld=0)
    #  Convective adjustment
    conv = climlab.convection.ConvectiveAdjustment(name='Convection',
                                                   state=state,
                                                   adj_lapse_rate=6.5)
    #  Couple everything together
    rcm = climlab.couple([rad,h2o,conv], name='Radiative-Convective Model')

    rcm.integrate_years(5.)
    assert np.abs(rcm.ASR - rcm.OLR) < 0.1  # close to energy balance
    rcm2 = climlab.process_like(rcm)
    rcm2.subprocess['Radiation'].absorber_vmr['CO2'] *= 2.
    rcm2.compute_diagnostics()
    assert (rcm2.ASR - rcm2.OLR) > 1.  # positive radiative forcing
    #  Test the xarray interface
    to_xarray(rcm2)

@pytest.mark.compiled
@pytest.mark.slow
def test_latitude():
    '''
    Run a radiative equilibrum model with RRTMG radiation out to equilibrium
    with an annual mean insolation profile as a function of latitude.
    '''
    num_lat = 8
    #  State variables (Air and surface temperature)
    state = climlab.column_state(num_lev=30, num_lat=num_lat, water_depth=1.)
    #  insolation
    #sol = climlab.radiation.AnnualMeanInsolation(domains=model.Ts.domain)
    sol = climlab.radiation.AnnualMeanInsolation(name='Insolation',
                                                 domains=state.Ts.domain)
    #  radiation module with insolation as input
    #   Set icld=0 for clear-sky only (no need to call cloud overlap routine)
    rad = climlab.radiation.RRTMG(name='Radiation', state=state, icld=0,
                                  S0=sol.S0,
                                  insolation=sol.insolation,
                                  coszen=sol.coszen)
    #  Couple everything together
    model = rad + sol
    #  Run out to equilibrium
    model.integrate_years(2.)
    #  Test for energy balance
    assert np.all(np.abs(model.ASR - model.OLR) < 0.1)
    #  Test for reasonable surface temperature gradient
    #  reversal of gradient at equator
    grad = np.diff(model.Ts, axis=0)
    assert np.all(grad[0:(int(num_lat/2)-1)] > 0.)
    assert np.all(grad[int(num_lat/2):] < 0.)

@pytest.mark.compiled
@pytest.mark.fast
def test_no_ozone():
    '''When user gives None as the ozone_file, the model is initialized
    with zero ozone. This should work on arbitrary grids.'''
    ps = 1060.
    num_lev=4000
    state = climlab.column_state(num_lev=num_lev, num_lat=1, water_depth=5.)
    lev = state.Tatm.domain.lev
    lev.bounds = np.linspace(0., ps, num_lev+1)
    lev.points = lev.bounds[:-1] + np.diff(lev.bounds)/2.
    lev.delta = np.abs(np.diff(lev.bounds))
    #  Create a RRTM radiation model
    rad = climlab.radiation.RRTMG(state=state, ozone_file=None)
    assert np.all(rad.absorber_vmr['O3']==0.)

@pytest.mark.compiled
@pytest.mark.fast
def test_fixed_insolation():
    '''Make sure that we can run a model forward with specified time-invariant insolation'''
    num_lat = 4; num_lev = 20   # grid size
    day_of_year = 80.  # days since Jan 1
    lat = np.linspace(-80., 80., num_lat)
    state = climlab.column_state(num_lev=num_lev, lat=lat)
    insolation = climlab.solar.insolation.daily_insolation(lat=lat, day=day_of_year)
    ins_array = insolation.values
    rad = climlab.radiation.RRTMG(name='Radiation', state=state, insolation=ins_array)
    rad.step_forward()

@pytest.mark.compiled
@pytest.mark.fast
def test_large_grid():
    num_lev = 50; num_lat=90
    state = climlab.column_state(num_lev=num_lev, num_lat=num_lat, water_depth=10.)
    rad1 = climlab.radiation.CAM3(state=state)
    rad1.step_forward()
    rad2 = climlab.radiation.RRTMG(state=state)
    rad2.step_forward()

    # Spectral OLR test
    # check that the spectrally decomposed TOA flux adds up to the normal OLR output
    rad3 = climlab.radiation.RRTMG(state=state, return_spectral_olr=True)
    rad3.step_forward()
    assert np.all(np.abs(rad3.OLR - rad3.OLR_spectral.sum(axis=-1))<0.1)
    #  Test the xarray interface
    to_xarray(rad3)




#==============================================================================
# Principles of the new `climlab` API design:
#
#     * `climlab.Process` object has several iterable dictionaries of named,
#       gridded variables:
#
#         * `process.state`
#
#             * state variables, usually time-dependent
#
#         - `process.input`
#             - boundary conditions and other gridded quantities independent of the
#             `process`
#             - often set by a parent `process`
#         - `process.param`  (which are basically just scalar `input`)
#         - `process.tendencies`
#             - iterable `dict` of time-tendencies (d/dt) for each state variable
#         - `process.diagnostics`
#             - any quantity derived from current state
# - The `process` is fully described by contents of `state`, `input` and `param`
# dictionaries. `tendencies` and `diagnostics` are always computable from current
# state.
# - `climlab` will remain (as much as possible) agnostic about the data formats
#     - Variables within the dictionaries will behave as `numpy.ndarray` objects
#     - Grid information and other domain details accessible as attributes
#     of each variable
#         - e.g. Tatm.lat
#         - Shortcuts like `process.lat` will work where these are unambiguous
# - Many variables will be accessible as process attributes `process.name`
#     - this restricts to unique field names in the above dictionaries
# - There may be other dictionaries that do have name conflicts
#     - e.g. dictionary of tendencies, with same keys as `process.state`
#     - These will *not* be accessible as `process.name`
#     - but *will* be accessible as `process.dict_name.name`
#     (as well as regular dict interface)
# - There will be a dictionary of named subprocesses `process.subprocess`
# - Each item in subprocess dict will itself be a `climlab.Process` object
# - For convenience with interactive work, each subprocess should be accessible
# as `process.subprocess.name` as well as `process.subprocess['name']`
# - `process.compute()` is a method that computes tendencies (d/dt)
#     - returns a dictionary of tendencies for all state variables
#     - keys for this dictionary are same as keys of state dictionary
#     - tendency dictionary is the total tendency including all subprocesses
#     - method only computes d/dt, does not apply changes
#     - thus method is relatively independent of numerical scheme
#         - may need to make exception for implicit scheme?
#     - method *will* update variables in `process.diagnostic`
#         - will also *gather all diagnostics* from `subprocesses`
# - `process.step_forward()` updates the state variables
#     - calls `process.compute()` to get current tendencies
#     - implements a particular time-stepping scheme
#     - user interface is agnostic about numerical scheme
# - `process.integrate_years()` etc will automate time-stepping
#     - also computation of time-average diagnostics.
# - Every `subprocess` should work independently of its parent `process` given
# appropriate `input`.
#     - investigating an individual `process` (possibly with its own
#     `subprocesses`) isolated from its parent needs to be as simple as doing:
#         - `newproc = climlab.process_like(procname.subprocess['subprocname'])`
#
#         - `newproc.compute()`
#         - anything in the `input` dictionary of `subprocname` will remain fixed
#==============================================================================

from __future__ import division, print_function
from builtins import object
import time, copy
import numpy as np
from climlab.domain.field import Field
from climlab.domain.domain import _Domain, zonal_mean_surface
from climlab.utils import walk
from climlab.utils.attrdict import AttrDict
from climlab.domain.xarray import state_to_xarray


def _make_dict(arg, argtype):
    if arg is None:
        return {}
    elif isinstance(arg, dict):
        return arg
    elif isinstance(arg, argtype):
        return {'default': arg}
    else:
        raise ValueError('Problem with input type')


class Process(object):
    """A generic parent class for all climlab process objects.
    Every process object has a set of state variables on a spatial grid.

    For more general information about `Processes` and their role in climlab,
    see :ref:`process_architecture` section climlab-architecture.

    **Initialization parameters** \n

    An instance of ``Process`` is initialized with the following
    arguments *(for detailed information see Object attributes below)*:

    :param Field state: spatial state variable for the process.
                        Set to ``None`` if not specified.
    :param domains:     domain(s) for the process
    :type domains:      :class:`~climlab.domain.domain._Domain` or dict of
                        :class:`~climlab.domain.domain._Domain`
    :param subprocess:  subprocess(es) of the process
    :type subprocess:   :class:`~climlab.process.process.Process` or dict of
                        :class:`~climlab.process.process.Process`
    :param array lat:   latitudinal points (optional)
    :param lev:         altitudinal points (optional)
    :param int num_lat: number of latitudional points (optional)
    :param int num_levels:
                        number of altitudinal points (optional)
    :param dict input:  collection of input quantities
    :param bool verbose: Flag to control text output during instantiation
                         of the Process [default: True]

    **Object attributes** \n

    Additional to the parent class :class:`~climlab.process.process.Process`
    following object attributes are generated during initialization:

    :ivar dict domains:     dictionary of process :class:`~climlab.domain.domain._Domain`
    :ivar dict state:       dictionary of process states
                            (of type :class:`~climlab.domain.field.Field`)
    :ivar dict param:       dictionary of model parameters which are given
                            through ``**kwargs``
    :ivar dict diagnostics: a dictionary with all diagnostic variables
    :ivar dict _input_vars: collection of input quantities like boundary conditions
                            and other gridded quantities
    :ivar str creation_date:
                            date and time when process was created
    :ivar subprocess:       dictionary of suprocesses of the process
    :vartype subprocess:    dict of :class:`~climlab.process.process.Process`

    """

    def __str__(self):
        str1 = 'climlab Process of type {0}. \n'.format(type(self))
        str1 += 'State variables and domain shapes: \n'
        for varname in list(self.state.keys()):
            str1 += '  {0}: {1} \n'.format(varname, self.domains[varname].shape)
        str1 += 'The subprocess tree: \n'
        str1 += walk.process_tree(self, name=self.name)
        return str1

    def __init__(self, name='Untitled', state=None, domains=None, subprocess=None,
                 lat=None, lev=None, num_lat=None, num_levels=None,
                 input=None, verbose=True, **kwargs):
        # verbose flag used to control text output at process creation time
        self.verbose = verbose
        self.name = name
        # dictionary of domains. Keys are the domain names
        self.domains = _make_dict(domains, _Domain)
        #  If lat is given, create a simple domains
        if lat is not None:
            sfc = zonal_mean_surface()
            self.domains.update({'default': sfc})
        # dictionary of state variables (all of type Field)
        self.state = AttrDict()
        states = _make_dict(state, Field)
        for name, value in states.items():
            self.set_state(name, value)
        # dictionary of model parameters
        self.param = kwargs
        # dictionary of diagnostic quantities
        #self.diagnostics = AttrDict()
        #self._diag_vars = frozenset()
        self._diag_vars = []
        # dictionary of input quantities
        #self.input = _make_dict(input, Field)
        if input is None:
            #self._input_vars = frozenset()
            self._input_vars = []
        else:
            self.add_input(list(input.keys()))
            for name, var in input:
                self.__dict__[name] = var
        self.creation_date = time.strftime("%a, %d %b %Y %H:%M:%S %z",
                                           time.localtime())
        # subprocess is a dictionary of any sub-processes
        self.subprocess = AttrDict()
        if subprocess is not None:
            self.add_subprocesses(subprocess)
        #if subprocess is None:
        #    #self.subprocess = {}
        #    # a dictionary whose items can be accessed as attributes
        #    self.subprocess = AttrDict()
        #else:
        #    self.add_subprocesses(subprocess)

    def add_subprocesses(self, procdict):
        """Adds a dictionary of subproceses to this process.

        Calls :func:`add_subprocess` for every process given in the
        input-dictionary. It can also pass a single process, which will
        be given the name *default*.

        :param procdict:    a dictionary with process names as keys
        :type procdict:     dict

        """
        if isinstance(procdict, Process):
            try:
                name = procdict.name
            except:
                name = 'default'
            self.add_subprocess(name, procdict)
        else:
            for name, proc in procdict.items():
                self.add_subprocess(name, proc)

    def add_subprocess(self, name, proc):
        """Adds a single subprocess to this process.

        :param string name:     name of the subprocess
        :param proc:            a Process object
        :type proc:             :class:`~climlab.process.process.Process`
        :raises: :exc:`ValueError`
                                if ``proc`` is not a process

        :Example:

            Replacing an albedo subprocess through adding a subprocess with
            same name::

                >>> from climlab.model.ebm import EBM_seasonal
                >>> from climlab.surface.albedo import StepFunctionAlbedo

                >>> # creating EBM model
                >>> ebm_s = EBM_seasonal()

                >>> print ebm_s

            .. code-block:: none
                :emphasize-lines: 8

                climlab Process of type <class 'climlab.model.ebm.EBM_seasonal'>.
                State variables and domain shapes:
                  Ts: (90, 1)
                The subprocess tree:
                top: <class 'climlab.model.ebm.EBM_seasonal'>
                   diffusion: <class 'climlab.dynamics.diffusion.MeridionalDiffusion'>
                   LW: <class 'climlab.radiation.AplusBT.AplusBT'>
                   albedo: <class 'climlab.surface.albedo.P2Albedo'>
                   insolation: <class 'climlab.radiation.insolation.DailyInsolation'>

            ::

                >>> #  creating and adding albedo feedback subprocess
                >>> step_albedo = StepFunctionAlbedo(state=ebm_s.state, **ebm_s.param)
                >>> ebm_s.add_subprocess('albedo', step_albedo)
                >>>
                >>> print ebm_s

            .. code-block:: none
                :emphasize-lines: 8

                climlab Process of type <class 'climlab.model.ebm.EBM_seasonal'>.
                State variables and domain shapes:
                  Ts: (90, 1)
                The subprocess tree:
                top: <class 'climlab.model.ebm.EBM_seasonal'>
                   diffusion: <class 'climlab.dynamics.diffusion.MeridionalDiffusion'>
                   LW: <class 'climlab.radiation.AplusBT.AplusBT'>
                   albedo: <class 'climlab.surface.albedo.StepFunctionAlbedo'>
                      iceline: <class 'climlab.surface.albedo.Iceline'>
                      cold_albedo: <class 'climlab.surface.albedo.ConstantAlbedo'>
                      warm_albedo: <class 'climlab.surface.albedo.P2Albedo'>
                   insolation: <class 'climlab.radiation.insolation.DailyInsolation'>

        """
        if isinstance(proc, Process):
            self.subprocess.update({name: proc})
            self.has_process_type_list = False
            # Add subprocess diagnostics to parent
            #  (if there are no name conflicts)
            for diagname, value in proc.diagnostics.items():
                #if not (diagname in self.diagnostics or hasattr(self, diagname)):
                #    self.add_diagnostic(diagname, value)
                self.add_diagnostic(diagname, value)
        else:
            raise ValueError('subprocess must be Process object')

    def remove_subprocess(self, name, verbose=True):
        """Removes a single subprocess from this process.

        :param string name:     name of the subprocess
        :param bool verbose:    information whether warning message
                                should be printed [default: True]

        :Example:

            Remove albedo subprocess from energy balance model::

                >>> import climlab
                >>> model = climlab.EBM()

                >>> print model
                climlab Process of type <class 'climlab.model.ebm.EBM'>.
                State variables and domain shapes:
                  Ts: (90, 1)
                The subprocess tree:
                top: <class 'climlab.model.ebm.EBM'>
                   diffusion: <class 'climlab.dynamics.diffusion.MeridionalDiffusion'>
                   LW: <class 'climlab.radiation.AplusBT.AplusBT'>
                   albedo: <class 'climlab.surface.albedo.StepFunctionAlbedo'>
                      iceline: <class 'climlab.surface.albedo.Iceline'>
                      cold_albedo: <class 'climlab.surface.albedo.ConstantAlbedo'>
                      warm_albedo: <class 'climlab.surface.albedo.P2Albedo'>
                   insolation: <class 'climlab.radiation.insolation.P2Insolation'>

                >>> model.remove_subprocess('albedo')

                >>> print model
                climlab Process of type <class 'climlab.model.ebm.EBM'>.
                State variables and domain shapes:
                  Ts: (90, 1)
                The subprocess tree:
                top: <class 'climlab.model.ebm.EBM'>
                   diffusion: <class 'climlab.dynamics.diffusion.MeridionalDiffusion'>
                   LW: <class 'climlab.radiation.AplusBT.AplusBT'>
                   insolation: <class 'climlab.radiation.insolation.P2Insolation'>

        """
        try:
            self.subprocess.pop(name)
        except KeyError:
            if verbose:
                print('WARNING: {} not found in subprocess dictionary.'.format(name))
        self.has_process_type_list = False

    def set_state(self, name, value):
        """Sets the variable ``name`` to a new state ``value``.

        :param string name:     name of the state
        :param value:           state variable
        :type value:            :class:`~climlab.domain.field.Field` or *array*
        :raises: :exc:`ValueError`
                                if state variable ``value`` is not having a domain.
        :raises: :exc:`ValueError`
                                if shape mismatch between existing domain and
                                new state variable.

        :Example:

            Resetting the surface temperature of an EBM to
            :math:`-5 ^{\circ} \\textrm{C}` on all latitues::

                >>> import climlab
                >>> from climlab import Field
                >>> import numpy as np

                >>> # setup model
                >>> model = climlab.EBM(num_lat=36)

                >>> # create new temperature distribution
                >>> initial = -5 * ones(size(model.lat))
                >>> model.set_state('Ts', Field(initial, domain=model.domains['Ts']))

                >>> np.squeeze(model.Ts)
                Field([-5., -5., -5., -5., -5., -5., -5., -5., -5., -5., -5., -5., -5.,
                       -5., -5., -5., -5., -5., -5., -5., -5., -5., -5., -5., -5., -5.,
                       -5., -5., -5., -5., -5., -5., -5., -5., -5., -5.])

        """
        if isinstance(value, Field):
            # populate domains dictionary with domains from state variables
            self.domains.update({name: value.domain})
        else:
            try:
                thisdom = self.state[name].domain
                domshape = thisdom.shape
            except:
                raise ValueError('State variable needs a domain.')
            value = np.atleast_1d(value)
            if value.shape == domshape:
                value = Field(value, domain=thisdom)
            else:
                raise ValueError('Shape mismatch between existing domain and new state variable.')
        # set the state dictionary
        self.state[name] = value
        for name, value in self.state.items():
        #convert int dtype to float
            if np.issubdtype(self.state[name].dtype, np.dtype('int').type):
                value = self.state[name].astype(float)
                self.state[name]=value
        self.__setattr__(name, value)

    def _guess_state_domains(self):
        for name, value in self.state.items():
            for domname, dom in self.domains.items():
                if value.shape == dom.shape:
                    # same shape, assume it's the right domain
                    self.state_domain[name] = dom

    def _add_field(self, field_type, name, value):
        """Adds a new field to a specified dictionary. The field is also added
        as a process attribute. field_type can be 'input', 'diagnostics' """
        try:
            self.__getattribute__(field_type).update({name: value})
        except:
            raise ValueError('Problem with field_type %s'  %field_type)
        #  Note that if process has attribute name, this will trigger The
        # setter method for that attribute
        self.__setattr__(name, value)

    def add_diagnostic(self, name, value=None):
        """Create a new diagnostic variable called ``name`` for this process
        and initialize it with the given ``value``.

        Quantity is accessible in two ways:

            * as a process attribute, i.e. ``proc.name``
            * as a member of the diagnostics dictionary,
              i.e. ``proc.diagnostics['name']``

        Use attribute method to set values, e.g.
        ```proc.name = value ```

        :param str name:        name of diagnostic quantity to be initialized
        :param array value:     initial value for quantity [default: None]

        :Example:

            Add a diagnostic CO2 variable to an energy balance model::

                >>> import climlab
                >>> model = climlab.EBM()

                >>> # initialize CO2 variable with value 280 ppm
                >>> model.add_diagnostic('CO2',280.)

                >>> # access variable directly or through diagnostic dictionary
                >>> model.CO2
                280
                >>> model.diagnostics.keys()
                ['ASR', 'CO2', 'net_radiation', 'icelat', 'OLR', 'albedo']

        """
        self._diag_vars.append(name)
        self.__setattr__(name, value)

    def add_input(self, name, value=None):
        '''Create a new input variable called ``name`` for this process
        and initialize it with the given ``value``.

        Quantity is accessible in two ways:

            * as a process attribute, i.e. ``proc.name``
            * as a member of the input dictionary,
              i.e. ``proc.input['name']``

        Use attribute method to set values, e.g.
        ```proc.name = value ```

        :param str name:        name of diagnostic quantity to be initialized
        :param array value:     initial value for quantity [default: None]
        '''
        self._input_vars.append(name)
        self.__setattr__(name, value)

    def declare_input(self, inputlist):
        '''Add the variable names in ``inputlist`` to the list of necessary inputs.'''
        for name in inputlist:
            self._input_vars.append(name)

    def declare_diagnostics(self, diaglist):
        '''Add the variable names in ``inputlist`` to the list of diagnostics.'''
        for name in diaglist:
            self._diag_vars.append(name)

    def remove_diagnostic(self, name):
        """	Removes a diagnostic from the ``process.diagnostic`` dictionary
        and also delete the associated process attribute.

        :param str name:    name of diagnostic quantity to be removed

        :Example:

            Remove diagnostic variable 'icelat' from energy balance model::

                >>> import climlab
                >>> model = climlab.EBM()

                >>> # display all diagnostic variables
                >>> model.diagnostics.keys()
                ['ASR', 'OLR', 'net_radiation', 'albedo', 'icelat']

                >>> model.remove_diagnostic('icelat')
                >>> model.diagnostics.keys()
                ['ASR', 'OLR', 'net_radiation', 'albedo']

                >>> # Watch out for subprocesses that may still want
                >>>  # to access the diagnostic 'icelat' variable !!!

        """
        #_ = self.diagnostics.pop(name)
        #delattr(type(self), name)
        try:
            delattr(self, name)
            self._diag_vars.remove(name)
        except:
            print('No diagnostic named {} was found.'.format(name))

    def to_xarray(self, diagnostics=False):
        """ Convert process variables to ``xarray.Dataset`` format.

        With ``diagnostics=True``, both state and diagnostic variables are included.

        Otherwise just the state variables are included.

        Returns an ``xarray.Dataset`` object with all spatial axes,
        including 'bounds' axes indicating cell boundaries in each spatial dimension.

        :Example:

            Create a single column radiation model and view as ``xarray`` object::

                >>> import climlab
                >>> state = climlab.column_state(num_lev=20)
                >>> model = climlab.radiation.RRTMG(state=state)

                >>> # display model state as xarray:
                >>> model.to_xarray()
                <xarray.Dataset>
                Dimensions:       (depth: 1, depth_bounds: 2, lev: 20, lev_bounds: 21)
                Coordinates:
                  * depth         (depth) float64 0.5
                  * depth_bounds  (depth_bounds) float64 0.0 1.0
                  * lev           (lev) float64 25.0 75.0 125.0 175.0 225.0 275.0 325.0 ...
                  * lev_bounds    (lev_bounds) float64 0.0 50.0 100.0 150.0 200.0 250.0 ...
                Data variables:
                    Ts            (depth) float64 288.0
                    Tatm          (lev) float64 200.0 204.1 208.2 212.3 216.4 220.5 224.6 ...

                >>> # take a single timestep to populate the diagnostic variables
                >>> model.step_forward()
                >>> # Now look at the full output in xarray format
                >>> model.to_xarray(diagnostics=True)
                <xarray.Dataset>
                Dimensions:           (depth: 1, depth_bounds: 2, lev: 20, lev_bounds: 21)
                Coordinates:
                  * depth             (depth) float64 0.5
                  * depth_bounds      (depth_bounds) float64 0.0 1.0
                  * lev               (lev) float64 25.0 75.0 125.0 175.0 225.0 275.0 325.0 ...
                  * lev_bounds        (lev_bounds) float64 0.0 50.0 100.0 150.0 200.0 250.0 ...
                Data variables:
                    Ts                (depth) float64 288.7
                    Tatm              (lev) float64 201.3 204.0 208.0 212.0 216.1 220.2 ...
                    ASR               (depth) float64 240.0
                    ASRcld            (depth) float64 0.0
                    ASRclr            (depth) float64 240.0
                    LW_flux_down      (lev_bounds) float64 0.0 12.63 19.47 26.07 32.92 40.1 ...
                    LW_flux_down_clr  (lev_bounds) float64 0.0 12.63 19.47 26.07 32.92 40.1 ...
                    LW_flux_net       (lev_bounds) float64 240.1 231.2 227.6 224.1 220.5 ...
                    LW_flux_net_clr   (lev_bounds) float64 240.1 231.2 227.6 224.1 220.5 ...
                    LW_flux_up        (lev_bounds) float64 240.1 243.9 247.1 250.2 253.4 ...
                    LW_flux_up_clr    (lev_bounds) float64 240.1 243.9 247.1 250.2 253.4 ...
                    LW_sfc            (depth) float64 128.9
                    LW_sfc_clr        (depth) float64 128.9
                    OLR               (depth) float64 240.1
                    OLRcld            (depth) float64 0.0
                    OLRclr            (depth) float64 240.1
                    SW_flux_down      (lev_bounds) float64 341.3 323.1 318.0 313.5 309.5 ...
                    SW_flux_down_clr  (lev_bounds) float64 341.3 323.1 318.0 313.5 309.5 ...
                    SW_flux_net       (lev_bounds) float64 240.0 223.3 220.2 217.9 215.9 ...
                    SW_flux_net_clr   (lev_bounds) float64 240.0 223.3 220.2 217.9 215.9 ...
                    SW_flux_up        (lev_bounds) float64 101.3 99.88 97.77 95.64 93.57 ...
                    SW_flux_up_clr    (lev_bounds) float64 101.3 99.88 97.77 95.64 93.57 ...
                    SW_sfc            (depth) float64 163.8
                    SW_sfc_clr        (depth) float64 163.8
                    TdotLW            (lev) float64 -1.502 -0.6148 -0.5813 -0.6173 -0.6426 ...
                    TdotLW_clr        (lev) float64 -1.502 -0.6148 -0.5813 -0.6173 -0.6426 ...
                    TdotSW            (lev) float64 2.821 0.5123 0.3936 0.3368 0.3174 0.3299 ...
                    TdotSW_clr        (lev) float64 2.821 0.5123 0.3936 0.3368 0.3174 0.3299 ...

        """
        if diagnostics:
            dic = self.state.copy()
            dic.update(self.diagnostics)
            return state_to_xarray(dic)
        else:
            return state_to_xarray(self.state)

    @property
    def diagnostics(self):
        """Dictionary access to all diagnostic variables

        :type:      dict

        """
        diag_dict = {}
        for key in self._diag_vars:
            try:
                #diag_dict[key] = getattr(self,key)
                #  using self.__dict__ doesn't count diagnostics defined as properties
                diag_dict[key] = self.__dict__[key]
            except:
                pass
        return diag_dict
    @property
    def input(self):
        """Dictionary access to all input variables

        That can be boundary conditions and other gridded quantities
        independent of the `process`

        :type:      dict

        """
        input_dict = {}
        for key in self._input_vars:
            try:
                input_dict[key] = getattr(self,key)
            except:
                pass
        return input_dict

    # Some handy shortcuts... only really make sense when there is only
    # a single axis of that type in the process.
    @property
    def lat(self):
        """Latitude of grid centers (degrees North)

        :getter:    Returns the points of axis ``'lat'`` if availible in the
                    process's domains.
        :type:      array
        :raises: :exc:`ValueError`
                    if no ``'lat'`` axis can be found.

        """
        try:
            for domname, dom in self.domains.items():
                try:
                    thislat = dom.axes['lat'].points
                except:
                    pass
            return thislat
        except:
            raise ValueError('Can\'t resolve a lat axis.')
    @property
    def lat_bounds(self):
        """Latitude of grid interfaces (degrees North)

        :getter:    Returns the bounds of axis ``'lat'`` if availible in the
                    process's domains.
        :type:      array
        :raises: :exc:`ValueError`
                    if no ``'lat'`` axis can be found.

        """
        try:
            for domname, dom in self.domains.items():
                try:
                    thislat = dom.axes['lat'].bounds
                except:
                    pass
            return thislat
        except:
            raise ValueError('Can\'t resolve a lat axis.')
    @property
    def lon(self):
        """Longitude of grid centers (degrees)

        :getter:    Returns the points of axis ``'lon'`` if availible in the
                    process's domains.
        :type:      array
        :raises: :exc:`ValueError`
                    if no ``'lon'`` axis can be found.

        """
        try:
            for domname, dom in self.domains.items():
                try:
                    thislon = dom.axes['lon'].points
                except:
                    pass
            return thislon
        except:
            raise ValueError('Can\'t resolve a lon axis.')
    @property
    def lon_bounds(self):
        """Longitude of grid interfaces (degrees)

        :getter:    Returns the bounds of axis ``'lon'`` if availible in the
                    process's domains.
        :type:      array
        :raises: :exc:`ValueError`
                    if no ``'lon'`` axis can be found.

        """
        try:
            for domname, dom in self.domains.items():
                try:
                    thislon = dom.axes['lon'].bounds
                except:
                    pass
            return thislon
        except:
            raise ValueError('Can\'t resolve a lon axis.')
    @property
    def lev(self):
        """Pressure levels at grid centers (hPa or mb)

        :getter:    Returns the points of axis ``'lev'`` if availible in the
                    process's domains.
        :type:      array
        :raises: :exc:`ValueError`
                    if no ``'lev'`` axis can be found.

        """
        try:
            for domname, dom in self.domains.items():
                try:
                    thislev = dom.axes['lev'].points
                except:
                    pass
            return thislev
        except:
            raise ValueError('Can\'t resolve a lev axis.')
    @property
    def lev_bounds(self):
        """Pressure levels at grid interfaces (hPa or mb)

        :getter:    Returns the bounds of axis ``'lev'`` if availible in the
                    process's domains.
        :type:      array
        :raises: :exc:`ValueError`
                    if no ``'lev'`` axis can be found.

        """
        try:
            for domname, dom in self.domains.items():
                try:
                    thislev = dom.axes['lev'].bounds
                except:
                    pass
            return thislev
        except:
            raise ValueError('Can\'t resolve a lev axis.')
    @property
    def depth(self):
        """Depth at grid centers (m)

        :getter:    Returns the points of axis ``'depth'`` if availible in the
                    process's domains.
        :type:      array
        :raises: :exc:`ValueError`
                    if no ``'depth'`` axis can be found.

        """
        try:
            for domname, dom in self.domains.items():
                try:
                    thisdepth = dom.axes['depth'].points
                except:
                    pass
            return thisdepth
        except:
            raise ValueError('Can\'t resolve a depth axis.')
    @property
    def depth_bounds(self):
        """Depth at grid interfaces (m)

        :getter:    Returns the bounds of axis ``'depth'`` if availible in the
                    process's domains.
        :type:      array
        :raises: :exc:`ValueError`
                    if no ``'depth'`` axis can be found.

        """
        try:
            for domname, dom in self.domains.items():
                try:
                    thisdepth = dom.axes['depth'].bounds
                except:
                    pass
            return thisdepth
        except:
            raise ValueError('Can\'t resolve a depth axis.')


def process_like(proc):
    """Make an exact clone of a process, including state and all subprocesses.

    The creation date is updated.

    :param proc:    process
    :type proc:     :class:`~climlab.process.process.Process`
    :return:        new process identical to the given process
    :rtype:         :class:`~climlab.process.process.Process`

    :Example:

        ::

            >>> import climlab
            >>> from climlab.process.process import process_like

            >>> model = climlab.EBM()
            >>> model.subprocess.keys()
            ['diffusion', 'LW', 'albedo', 'insolation']

            >>> albedo = model.subprocess['albedo']
            >>> albedo_copy = process_like(albedo)

            >>> albedo.creation_date
            'Thu, 24 Mar 2016 01:32:25 +0000'

            >>> albedo_copy.creation_date
            'Thu, 24 Mar 2016 01:33:29 +0000'

    """
    newproc = copy.deepcopy(proc)
    newproc.creation_date = time.strftime("%a, %d %b %Y %H:%M:%S %z",
                                          time.localtime())
    return newproc


def get_axes(process_or_domain):
    """Returns a dictionary of all Axis in a domain or dictionary of domains.

    :param process_or_domain:   a process or a domain object
    :type process_or_domain:    :class:`~climlab.process.process.Process` or
                                :class:`~climlab.domain.domain._Domain`
    :raises: :exc:              `TypeError` if input is not or not having a domain
    :returns:                   dictionary of input's Axis
    :rtype:                     dict

    :Example:

        ::

            >>> import climlab
            >>> from climlab.process.process import get_axes

            >>> model = climlab.EBM()

            >>> get_axes(model)
            {'lat': <climlab.domain.axis.Axis object at 0x7ff13b9dd2d0>,
             'depth': <climlab.domain.axis.Axis object at 0x7ff13b9dd310>}

    """
    if isinstance(process_or_domain, Process):
        dom = process_or_domain.domains
    else:
        dom = process_or_domain
    if isinstance(dom, _Domain):
        return dom.axes
    elif isinstance(dom, dict):
        axes = {}
        for thisdom in list(dom.values()):
            assert isinstance(thisdom, _Domain)
            axes.update(thisdom.axes)
        return axes
    else:
        raise TypeError('dom must be a domain or dictionary of domains.')



from __future__ import absolute_import, division, print_function
from .time_dependent_process import TimeDependentProcess

class ExternalForcing(TimeDependentProcess):
    """A Process class for user-defined tendencies of state variables.
    Useful for combining some prescribed external forcing with an interactive model.

    :Example:
        The user can invoke the process on a dicionary of state variables ``mystate`` like this::

            myforcing = climlab.process.ExternalForcing(state=mystate)

        and then set the desired tendencies in the dictionary ``myforcing.forcing_tendencies``,
        in units of [state variable unit] per second.
    """
    def __init__(self,**kwargs):
        super(ExternalForcing, self).__init__(**kwargs)
        self.forcing_tendencies = {}
        for var in self.state:
            self.forcing_tendencies[var] = 0. * self.state[var]

    def _compute(self):
        return self.forcing_tendencies



'''
This package contains ready-made models that can be run "off-the-shelf".

    :Example:

        .. code-block:: python

            import climlab
            #  create a 1D Energy Balance Model
            mymodel = climlab.EBM()
            #  see what you just created
            print(mymodel)
            #  run the model
            mymodel.integrate_years(2.)
            #  display the current state
            mymodel.state
            #  see what diagnostics have been computed
            mymodel.diagnostics.keys()

These modules are fully functional and tested.
However users are encouraged to build their own models
by explicitly creating individual processes and coupling together
as subprocesses of a parent process.

See the documentation for the RRTMG scheme for an example of building a
radiative-convective column model from individual components.
'''



r"""General solver of the 1D meridional advection-diffusion equation on the sphere:

.. math::

    \frac{\partial}{\partial t} \psi(\phi,t) &= -\frac{1}{a \cos\phi} \frac{\partial}{\partial \phi} \left[ \cos\phi ~ F(\phi,t) \right] \\
    F &= U(\phi) \psi(\phi) -\frac{K(\phi)}{a} ~ \frac{\partial \psi}{\partial \phi}

for a state variable :math:`\psi(\phi,t)`, arbitrary diffusivity :math:`K(\phi)`
in units of :math:`x^2 ~ t^{-1}`, and advecting velocity :math:`U(\phi)`.
:math:`\phi` is latitude and :math:`a` is the Earth's radius (in meters).

:math:`K` and :math:`U` can be scalars,
or optionally vector *specified at grid cell boundaries*
(so their lengths must be exactly 1 greater than the length of :math:`\phi`).

:math:`K` and :math:`U` can be modified by the user at any time
(e.g., after each timestep, if they depend on other state variables).

A fully implicit timestep is used for computational efficiency. Thus the computed
tendency :math:`\frac{\partial \psi}{\partial t}` will depend on the timestep.

In addition to the tendency over the implicit timestep,
the solver also calculates several diagnostics from the updated state:

- ``diffusive_flux`` given by :math:`-\frac{K(\phi)}{a} ~ \frac{\partial \psi}{\partial \phi}` in units of :math:`[\psi]~[x]`/s
- ``advective_flux`` given by :math:`U(\phi) \psi(\phi)` (same units)
- ``total_flux``, the sum of advective, diffusive and prescribed fluxes
- ``flux_convergence`` (or instantanous scalar tendency) given by the right hand side of the first equation above, in units of :math:`[\psi]`/s

Non-uniform grid spacing is supported.

The state variable :math:`\psi` may be multi-dimensional, but the diffusion
will operate along the latitude dimension only.
"""
from __future__ import division
import numpy as np
from .advection_diffusion import AdvectionDiffusion, Diffusion
from climlab import constants as const


class MeridionalAdvectionDiffusion(AdvectionDiffusion):
    """A parent class for meridional advection-diffusion processes.
    """
    def __init__(self,
                 K=0.,
                 U=0.,
                 use_banded_solver=False,
                 prescribed_flux=0.,
                 **kwargs):
        super(MeridionalAdvectionDiffusion, self).__init__(K=K, U=U,
                diffusion_axis='lat', use_banded_solver=use_banded_solver, **kwargs)
        # Conversion of delta from degrees (grid units) to physical length units
        phi_stag = np.deg2rad(self.lat_bounds)
        phi = np.deg2rad(self.lat)
        self._Xcenter[...,:] = phi*const.a
        self._Xbounds[...,:] = phi_stag*const.a
        self._weight_bounds[...,:] = np.cos(phi_stag)
        self._weight_center[...,:] = np.cos(phi)
        #  Now properly compute the weighted advection-diffusion matrix
        self.prescribed_flux = prescribed_flux
        self.K = K
        self.U = U


class MeridionalDiffusion(MeridionalAdvectionDiffusion):
    """A parent class for meridional diffusion-only processes,
    with advection set to zero.

    Otherwise identical to the parent class.
    """
    def __init__(self,
                 K=0.,
                 use_banded_solver=False,
                 prescribed_flux=0.,
                 **kwargs):
        # Just initialize the AdvectionDiffusion class with U=0
        super(MeridionalDiffusion, self).__init__(
            U=0.,
            K=K,
            prescribed_flux=prescribed_flux,
            use_banded_solver=use_banded_solver, **kwargs)



from __future__ import division
import numpy as np
import climlab
import pytest

@pytest.mark.fast
def test_state():
    initialT0 = 15.
    sfc = climlab.domain.surface_2D(num_lat=90, num_lon=180)
    sfc = climlab.domain.surface_2D(lat=([-90.,0.,90.]),
                            lon=([-180.,0.,180.]))
    state = climlab.surface_state(T0 = initialT0, num_lat=90, num_lon=180)
    assert state.Ts.ndim == 3
    assert state.Ts.shape == (90, 180, 1)
    assert np.isclose(climlab.global_mean(state.Ts), initialT0, atol=1E-02)

@pytest.mark.fast
def test_2D_EBM():
    '''Can we step forward a 2D lat/lon EBM?'''
    #state = climlab.surface_state(num_lon=4)
    #m = climlab.EBM_annual(state=state)
    m = climlab.EBM_annual(num_lon=4)
    m.step_forward()
    assert m.state.Ts.shape == (90, 4, 1)
    # Test the xarray interface
    m.to_xarray()

@pytest.mark.fast
def test_2D_EBM_seasonal():
    '''Can we step forward a 2D seasonal lat/lon EBM?'''
    #state = climlab.surface_state(num_lon=4)
    #m = climlab.EBM_annual(state=state)
    m = climlab.EBM_seasonal(num_lon=4)
    m.step_forward()
    assert m.state.Ts.shape == (90, 4, 1)
    # Test the xarray interface
    m.to_xarray()

@pytest.mark.fast
def test_2D_insolation():
    #state = climlab.surface_state(num_lon=4)
    #m = climlab.EBM_annual(state=state)
    m = climlab.EBM_annual(num_lon=4)
    #  the answers are the mean of 1D insolation arrays
    #  the mean shouldn't change from 1D to 2D...
    #  there are exactly the same amount of each number in 2D array
    assert np.mean(m.subprocess['insolation'].insolation) == pytest.approx(299.30467670961832)
    #assert np.mean(m.subprocess['insolation'].insolation) == 299.30467670961832
    sfc = m.domains['Ts']
    m.add_subprocess('insolation',
        climlab.radiation.P2Insolation(domains=sfc, **m.param))
    assert np.mean(m.subprocess['insolation'].insolation) == pytest.approx(300.34399999999999)
    #assert np.mean(m.subprocess['insolation'].insolation) == 300.34399999999999



from __future__ import division
import numpy as np
import climlab
import pytest
from climlab.tests.xarray_test import to_xarray


@pytest.fixture()
def model():
    return climlab.GreyRadiationModel(num_lev=30, num_lat=90)

@pytest.fixture()
def model_with_insolation(model):
    insolation = climlab.radiation.DailyInsolation(domains=model.Ts.domain)
    model.add_subprocess('insolation', insolation)
    model.subprocess.SW.flux_from_space = insolation.insolation
    return model

@pytest.fixture()
def rcmodel():
    model2 = climlab.RadiativeConvectiveModel(num_lev=30, num_lat=90)
    insolation = climlab.radiation.DailyInsolation(domains=model2.Ts.domain)
    model2.add_subprocess('insolation', insolation)
    model2.subprocess.SW.flux_from_space = insolation.insolation
    return model2

@pytest.fixture()
def diffmodel(rcmodel):
    diffmodel = climlab.process_like(rcmodel)
    # meridional diffusivity in m**2/s
    K = 0.05 / diffmodel.Tatm.domain.heat_capacity[0] *  climlab.constants.a**2
    d = climlab.dynamics.MeridionalDiffusion(K=K,
                state={'Tatm': diffmodel.state['Tatm']},
                **diffmodel.param)
    diffmodel.add_subprocess('diffusion', d)
    return diffmodel

# helper for a common test pattern
def _check_minmax(array, amin, amax):
    return (np.allclose(array.min(), amin) and
            np.allclose(array.max(), amax))

@pytest.mark.fast
def test_model_creation(model):
    """Just make sure we can create a model."""
    assert len(model.lat)==90
    # Test the xarray interface
    to_xarray(model)

@pytest.mark.fast
def test_add_insolation(model_with_insolation):
    """"Create a model with insolation and check that SW_down_TOA has
    reasonable values."""
    model_with_insolation.step_forward()
    assert _check_minmax(model_with_insolation.SW_down_TOA, 0, 555.17111)

@pytest.mark.slow
def test_integrate_years(model_with_insolation):
    """Check that we can integrate forward the model and get the expected
    surface temperature."""
    model_with_insolation.step_forward()
    model_with_insolation.integrate_years(1)
    ts = model_with_insolation.timeave['Ts']
    assert _check_minmax(ts, 225.402329962, 301.659494398)

@pytest.mark.slow
def test_rcmodel(rcmodel):
    """Check that we can integrate forwrd the radiative convective model and
    get expected atmospheric temperature."""
    rcmodel.step_forward()
    rcmodel.integrate_years(1)
    tatm = rcmodel.timeave['Tatm']
    assert _check_minmax(tatm, 176.786517491, 292.222277112)

@pytest.mark.slow
def test_diffmodel(diffmodel):
    """Check that we can integrate the model with diffusion."""
    diffmodel.step_forward()
    diffmodel.integrate_years(1)
    tatm = diffmodel.timeave['Tatm']
    assert _check_minmax(tatm, 208.689339823, 285.16085319)

@pytest.mark.fast
def test_external_tendency():
    """Check that we can add an externally defined tendency to a
    radiative-convective model."""
    model = climlab.GreyRadiationModel(num_lev=30)
    model2 = climlab.process_like(model)
    model.step_forward()
    ext = climlab.process.ExternalForcing(state=model2.state)
    temp_tend = 1E-5  # K/s
    ext.forcing_tendencies['Tatm'][:] = temp_tend
    model2.add_subprocess('External', ext)
    model2.step_forward()
    assert model.tendencies['Tatm'] + temp_tend == pytest.approx(model2.tendencies['Tatm'])
    #assert np.all(np.isclose(model.tendencies['Tatm'] == (model2.tendencies['Tatm']-temp_tend))



"""Convenience routines for setting up initial conditions."""
from __future__ import division
import numpy as np
from climlab.domain import domain
from climlab.domain.field import Field
from climlab.utils.attrdict import AttrDict
from climlab.utils import legendre


def column_state(num_lev=30,
                 num_lat=1,
                 lev=None,
                 lat=None,
                 water_depth=1.0):
    """Sets up a state variable dictionary consisting of temperatures
    for atmospheric column (``Tatm``) and surface mixed layer (``Ts``).

    Surface temperature is always 288 K. Atmospheric temperature is initialized
    between 278 K at lowest altitude and 200 at top of atmosphere according to
    the number of levels given.

    **Function-call arguments** \n

    :param int num_lev:         number of pressure levels
                                (evenly spaced from surface to top of atmosphere)
                                [default: 30]
    :param int num_lat:         number of latitude points on the axis
                                [default: 1]
    :param lev:                 specification for height axis (optional)
    :type lev:                  :class:`~climlab.domain.axis.Axis`
                                or pressure array
    :param array lat:           size of array determines dimension of latitude
                                (optional)
    :param float water_depth:   *irrelevant*

    :returns:                   dictionary with two temperature
                                :class:`~climlab.domain.field.Field`
                                for atmospheric column ``Tatm`` and
                                surface mixed layer ``Ts``
    :rtype:                     dict

    :Example:

        ::

            >>> from climlab.domain import initial
            >>> T_dict = initial.column_state()

            >>> print T_dict
            {'Tatm': Field([ 200.        ,  202.68965517,  205.37931034,  208.06896552,
                    210.75862069,  213.44827586,  216.13793103,  218.82758621,
                    221.51724138,  224.20689655,  226.89655172,  229.5862069 ,
                    232.27586207,  234.96551724,  237.65517241,  240.34482759,
                    243.03448276,  245.72413793,  248.4137931 ,  251.10344828,
                    253.79310345,  256.48275862,  259.17241379,  261.86206897,
                    264.55172414,  267.24137931,  269.93103448,  272.62068966,
                    275.31034483,  278.        ]), 'Ts': Field([ 288.])}

    """
    if lat is not None:
        num_lat = np.array(lat).size
    if lev is not None:
        num_lev = np.array(lev).size

    if num_lat == 1:
        sfc, atm = domain.single_column(water_depth=water_depth,
                                        num_lev=num_lev,
                                        lev=lev)
    else:
        sfc, atm = domain.zonal_mean_column(water_depth=water_depth,
                                            num_lev=num_lev,
                                            lev=lev,
                                            num_lat=num_lat,
                                            lat=lat)
    num_lev = atm.lev.num_points
    Ts = Field(288.*np.ones(sfc.shape), domain=sfc)
    Tinitial = np.tile(np.linspace(200., 288.-10., num_lev), sfc.shape)
    Tatm = Field(Tinitial, domain=atm)
    state = AttrDict()
    state['Ts'] = Ts
    state['Tatm'] = Tatm
    return state


def surface_state(num_lat=90,
                  num_lon=None,
                  water_depth=10.,
                  T0=12.,
                  T2=-40.):
    """Sets up a state variable dictionary for a surface model
    (e.g. :class:`~climlab.model.ebm.EBM`) with a uniform slab ocean depth.

    The domain is either 1D (latitude) or 2D (latitude, longitude)
    depending on whether the input argument num_lon is supplied.

    Returns a single state variable `Ts`, the temperature of the surface
    mixed layer (slab ocean).

    The temperature is initialized to a smooth equator-to-pole shape given by

    .. math::

        T(\phi) = T_0 + T_2 P_2(\sin\phi)

    where :math:`\phi` is latitude, and :math:`P_2` is the second Legendre
    polynomial :class:`~climlab.utils.legendre.P2`.

    **Function-call arguments** \n

    :param int num_lat:         number of latitude points [default: 90]
    :param int num_lat:         (optional) number of longitude points [default: None]
    :param float water_depth:   depth of the slab ocean in meters [default: 10.]
    :param float T0:            global-mean initial temperature in :math:`^{\circ} \\textrm{C}` [default: 12.]
    :param float T2:            2nd Legendre coefficient for equator-to-pole gradient in
                                initial temperature, in :math:`^{\circ} \\textrm{C}` [default: -40.]

    :returns:                   dictionary with temperature
                                :class:`~climlab.domain.field.Field`
                                for surface mixed layer ``Ts``
    :rtype:                     dict


    :Example:

        ::

            >>> from climlab.domain import initial
            >>> import numpy as np

            >>> T_dict = initial.surface_state(num_lat=36)

            >>> print np.squeeze(T_dict['Ts'])
            [-27.88584094 -26.97777479 -25.18923361 -22.57456133 -19.21320344
             -15.20729309 -10.67854785  -5.76457135  -0.61467228   4.61467228
               9.76457135  14.67854785  19.20729309  23.21320344  26.57456133
              29.18923361  30.97777479  31.88584094  31.88584094  30.97777479
              29.18923361  26.57456133  23.21320344  19.20729309  14.67854785
               9.76457135   4.61467228  -0.61467228  -5.76457135 -10.67854785
             -15.20729309 -19.21320344 -22.57456133 -25.18923361 -26.97777479
             -27.88584094]

    """
    if num_lon is None:
        sfc = domain.zonal_mean_surface(num_lat=num_lat,
                                        water_depth=water_depth)
    else:
        sfc = domain.surface_2D(num_lat=num_lat,
                                num_lon=num_lon,
                                water_depth=water_depth)
    if 'lon' in sfc.axes:
        lon, lat = np.meshgrid(sfc.axes['lon'].points, sfc.axes['lat'].points)
    else:
        lat = sfc.axes['lat'].points
    sinphi = np.sin(np.deg2rad(lat))
    initial = T0 + T2 * legendre.P2(sinphi)
    Ts = Field(initial, domain=sfc)
    #if num_lon is None:
    #    Ts = Field(initial, domain=sfc)
    #else:
    #    Ts = Field([[initial for k in range(num_lon)]], domain=sfc)
    state = AttrDict()
    state['Ts'] = Ts
    return state
